# -*- coding: utf-8 -*-
"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Tuple, List, Optional, Any, Union, Iterable, cast

from qt import Qt, QSize
from qt import QWidget, QHBoxLayout, QVBoxLayout, \
    QPushButton, QSpacerItem, QSizePolicy, \
    QTreeView, QFont, QAbstractItemView, QHeaderView
from qt import QModelIndex, QPersistentModelIndex
from kdestub import KApplication
from rule import Rule, Ruleset, PredefinedRuleset, ParameterRule, BoolRule
from util import uniqueList
from mi18n import i18n, i18nc, i18ncE, english
from differ import RulesetDiffer
from common import Debug
from tree import TreeItem, RootItem, TreeModel
from dialogs import Sorry
from modeltest import ModelTest
from genericdelegates import RightAlignedCheckboxDelegate, ZeroEmptyColumnDelegate
from statesaver import StateSaver
from guiutil import decorateWindow

if TYPE_CHECKING:
    from qt import QShowEvent, QHideEvent, QItemSelection, QObject, QLayout
    from rule import RuleList


class RuleRootItem(RootItem):

    """the root item for the ruleset tree"""

    def columnCount(self) ->int:
        return len(self.raw)


class RuleTreeItem(TreeItem):

    """generic class for items in our rule tree"""
    # pylint: disable=abstract-method
    # we know content() is abstract, this class is too

    def columnCount(self) ->int:
        """can be different for every rule"""
        if hasattr(self, 'colCount'):
            return self.colCount
        return len(self.raw)

    def ruleset(self) ->Ruleset:
        """return the ruleset containing this item"""
        item:TreeItem = self
        while not isinstance(item.raw, Ruleset):
            assert item.parent
            item = item.parent
        return item.raw

    def tooltip(self) ->str:
        """the tooltip"""
        return ""


class RulesetItem(RuleTreeItem):

    """represents a ruleset in the tree"""

    def __init__(self, content:Ruleset) ->None:
        super().__init__(content)

    @property
    def raw(self) ->Ruleset:
        return cast(Ruleset, super().raw)

    def content(self, column:int) ->str:
        """return content stored in this item"""
        if column == 0:
            return self.raw.name
        return ''

    def columnCount(self) ->int:
        return 1

    def remove(self) ->None:
        """remove this ruleset from the model and the database"""
        self.raw.remove()

    def tooltip(self) ->str:
        """the tooltip for a ruleset"""
        return self.raw.description


class RuleListItem(RuleTreeItem):

    """represents a list of rules in the tree"""

    def __init__(self, content:List[ParameterRule]) ->None:
        super().__init__(content)

    @property
    def raw(self) ->'RuleList':
        return cast('RuleList', super().raw)

    def content(self, column:int) ->str:
        """return content stored in this item"""
        if column == 0:
            return self.raw.name
        return ''

    def tooltip(self) ->str:
        """tooltip for a list item explaining the usage of this list"""
        ruleset = self.ruleset()
        return '<b>' + i18n(ruleset.name) + '</b><br><br>' + \
            i18n(self.raw.description)


class RuleItem(RuleTreeItem):

    """represents a rule in the tree"""

    def __init__(self, content:ParameterRule) ->None:
        super().__init__(content)

    @property
    def raw(self) ->Rule:
        return cast(Rule, super().raw)

    def content(self, column:int) ->str:
        """return the content stored in this node"""
        colNames = self.parent.parent.parent.raw  # type:ignore
        if column == 0:
            return self.raw.name
        if isinstance(self.raw, ParameterRule):
            if column == 1:
                return self.raw.parameter
        else:
            _ = str(column)
            if not hasattr(self.raw.score, _):
                _ = colNames[column]
            return getattr(self.raw.score, _)
        return ''

    def tooltip(self) ->str:
        """tooltip for rule: just the name of the ruleset"""
        ruleset = self.ruleset()
        if self.raw.description:
            return '<b>' + i18n(ruleset.name) + '</b><br><br>' + \
                i18n(self.raw.description)
        return i18n(ruleset.name)


class RuleModel(TreeModel):

    """a model for our rule table"""

    def __init__(self, rulesets:List[Ruleset], title:str, parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.rulesets = rulesets
        self.loaded = False
        unitPairs:List[Tuple[str, int]] = []
        # unitPairs: int is the priority:show 0 leftmost, 9999 rightmost
        for ruleset in rulesets:
            ruleset.load()
            for rule in ruleset.allRules:
                unitPairs.extend(rule.score.unitNames.items())
        unitPairs = sorted(unitPairs, key=lambda x: x[1])
        unitNames = uniqueList(x[0] for x in unitPairs)
        rootData = [title]
        rootData.extend(unitNames)
        self.rootItem = RuleRootItem(rootData)

    def canFetchMore(self, unusedParent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->bool:
        """did we already load the rules? We only want to do that
        when the config tab with rulesets is actually shown"""
        return not self.loaded

    def fetchMore(self, unusedParent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->None:
        """load the rules"""
        for ruleset in self.rulesets:
            self.appendRuleset(ruleset)
        self.loaded = True

    def data(self, index:Union[QModelIndex,QPersistentModelIndex], role:int=Qt.ItemDataRole.DisplayRole) ->Any:
        """get data fom model"""
        # pylint: disable=too-many-branches
        # too many branches
        result = None
        if index.isValid() and isinstance(index, QModelIndex):
            item = cast(RuleTreeItem, index.internalPointer())
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
                if index.column() == 1:
                    if isinstance(item, RuleItem) and isinstance(item.raw, BoolRule):
                        return ''
                showValue = item.content(index.column())
                if isinstance(showValue, str) and showValue.endswith('.0'):
                    try:
                        showValue = str(int(float(showValue)))
                    except ValueError:
                        pass
                if showValue == '0':
                    showValue = ''
                result = showValue
            elif role == Qt.ItemDataRole.CheckStateRole:
                if self.isCheckboxCell(index):
                    bData = item.content(index.column())
                    result = Qt.CheckState.Checked if bData else Qt.CheckState.Unchecked
            elif role == Qt.ItemDataRole.TextAlignmentRole:
                result = int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
                if index.column() > 0:
                    result = int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
            elif role == Qt.ItemDataRole.FontRole and index.column() == 0:
                ruleset = item.ruleset()
                if isinstance(ruleset, PredefinedRuleset):
                    font = QFont()
                    font.setItalic(True)
                    result = font
            elif role == Qt.ItemDataRole.ToolTipRole:
                assert isinstance(item, RuleTreeItem), item
                tip = f'<b></b>{i18n(item.tooltip())}<b></b>' if item else ''
                result = tip
        return result

    @staticmethod
    def isCheckboxCell(index:QModelIndex) ->bool:
        """are we dealing with a checkbox?"""
        if index.column() != 1:
            return False
        item = index.internalPointer()
        return isinstance(item, RuleItem) and isinstance(item.raw, BoolRule)

    def headerData(self, section:int, orientation:Qt.Orientation, role:int=Qt.ItemDataRole.DisplayRole) ->Optional[Any]:
        """tell the view about the wanted headers"""
        if Qt is None:
            # happens when kajongg exits unexpectedly
            return None
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            assert self.rootItem
            if section >= self.rootItem.columnCount():
                return None
            result = self.rootItem.content(section)
            if result == 'doubles':
                return 'x2'
            return i18nc('kajongg', result)
        if role == Qt.ItemDataRole.TextAlignmentRole:
            leftRight = Qt.AlignmentFlag.AlignLeft if section == 0 else Qt.AlignmentFlag.AlignRight
            return int(leftRight | Qt.AlignmentFlag.AlignVCenter)
        return None

    def appendRuleset(self, ruleset:Ruleset) ->None:
        """append ruleset to the model"""
        if not ruleset:
            return
        ruleset.load()
        parent = QModelIndex()
        assert self.rootItem
        row = self.rootItem.childCount()
        rulesetItems = list([RulesetItem(ruleset)])
        self.insertRows(row, rulesetItems, parent)
        rulesetIndex = self.index(row, 0, parent)
        ruleLists:List[List[ParameterRule]] = [x for x in ruleset.ruleLists if len(x)]
        ruleListItems:List[RuleListItem] = [RuleListItem(x) for x in ruleLists]
        for item in ruleListItems:
            item.colCount = self.rootItem.columnCount()  # type:ignore
        self.insertRows(0, ruleListItems, rulesetIndex)
        for ridx, ruleList in enumerate(ruleLists):
            listIndex = self.index(ridx, 0, rulesetIndex)
            ruleItems = [RuleItem(x) for x in ruleList if 'internal' not in x.options]
            self.insertRows(0, ruleItems, listIndex)


class EditableRuleModel(RuleModel):

    """add methods needed for editing"""

    def __setRuleData(self, column:int, content:Union[Rule, ParameterRule],
        value:Union[str, int, bool]) ->Tuple[bool, Optional[str]]:
        """change rule data in the model"""
        dirty, message = False, None
        if column == 1 and isinstance(content, ParameterRule):
            oldParameter = content.parameter
            if isinstance(content, BoolRule):
                return False, ''
            if content.parameter != value:
                dirty = True
                content.parameter = value
            message = content.validate()
            if message:
                content.parameter = oldParameter
                dirty = False
        else:
            assert self.rootItem
            unitName = self.rootItem.content(column)
            assert isinstance(value, (int, float))
            dirty, message = content.score.change(unitName, value)
        return dirty, message

    def setData(self, index:Union[QModelIndex,QPersistentModelIndex], value:str,
        role:int=Qt.ItemDataRole.EditRole) ->bool:
        """change data in the model"""
        # pylint:  disable=too-many-branches
        if not index.isValid() or not isinstance(index, QModelIndex):
            return False
        dirty = False
        column = index.column()
        item = cast(RuleTreeItem, index.internalPointer())
        ruleset = item.ruleset()
        content = item.raw
        if role == Qt.ItemDataRole.EditRole:
            if isinstance(content, Ruleset) and column == 0:
                oldName = content.name
                content.rename(english(value))
                dirty = oldName != content.name
            elif isinstance(content, (Rule, ParameterRule)):
                dirty, message = self.__setRuleData(column, content, value)
                if message:
                    Sorry(message)
                    return False
            else:
                return False
        elif role == Qt.ItemDataRole.CheckStateRole:
            if isinstance(content, BoolRule) and column == 1:
                if not isinstance(ruleset, PredefinedRuleset):
                    newValue = value == Qt.CheckState.Checked
                    if content.parameter != newValue:
                        dirty = True
                        content.parameter = newValue
            else:
                return False
        if dirty:
            if isinstance(content, (Rule, ParameterRule)):
                ruleset.updateRule(content)
            self.dataChanged.emit(index, index)
        return True

    def flags(self, index:QModelIndex) ->Qt.ItemFlag:  # type:ignore[override]
        """tell the view what it can do with this item"""
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        column = index.column()
        item = cast(RuleTreeItem, index.internalPointer())
        content = item.raw
        checkable = False
        if isinstance(content, Ruleset) and column == 0:
            mayEdit = True
        elif isinstance(content, (Rule, ParameterRule)):
            checkable = column == 1 and isinstance(content, BoolRule)
            mayEdit = bool(column)
        else:
            mayEdit = False
        mayEdit = mayEdit and not isinstance(item.ruleset(), PredefinedRuleset)
        result = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        if mayEdit:
            result |= Qt.ItemFlag.ItemIsEditable
        if checkable:
            result |= Qt.ItemFlag.ItemIsUserCheckable
        return cast(Qt.ItemFlag, result)


class RuleTreeView(QTreeView):

    """Tree view for our rulesets"""

    def __init__(self, name:str,
                 btnCopy:Optional[QPushButton]=None,
                 btnRemove:Optional[QPushButton]=None,
                 btnCompare:Optional[QPushButton]=None, parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.name = name
        self.setObjectName('RuleTreeView')
        self.btnCopy = btnCopy
        self.btnRemove = btnRemove
        self.btnCompare = btnCompare
        for button in [self.btnCopy, self.btnRemove, self.btnCompare]:
            if button:
                button.setEnabled(False)
        if header := self.header():
            header.setObjectName('RuleTreeViewHeader')
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.ruleModel:Optional[RuleModel] = None
        self.ruleModelTest = None
        self.rulesets:List[Ruleset] = []  # nasty: this generates self.ruleModel
        self.differs:List[RulesetDiffer] = []

    def dataChanged(self, unusedIndex1:Union[QModelIndex,QPersistentModelIndex],
        unusedIndex2:Union[QModelIndex,QPersistentModelIndex],
        unusedRoles:Optional[Iterable[int]]=None) ->None:
        """get called if the model has changed: Update all differs"""
        for differ in self.differs:
            differ.rulesetChanged()

    @property
    def rulesets(self) ->List[Ruleset]:
        """a list of rulesets made available by this model"""
        assert self.ruleModel
        return self.ruleModel.rulesets

    @rulesets.setter
    def rulesets(self, rulesets:List[Ruleset]) ->None:
        """a new list: update display"""
        if not self.ruleModel or self.ruleModel.rulesets != rulesets:
            if self.btnRemove and self.btnCopy:
                self.ruleModel = EditableRuleModel(rulesets, self.name)
            else:
                self.ruleModel = RuleModel(rulesets, self.name)
            delegate = RightAlignedCheckboxDelegate(self, self.ruleModel.isCheckboxCell)
            self.setItemDelegateForColumn(1, delegate)
            for  col in (2, 3):
                self.setItemDelegateForColumn(col, ZeroEmptyColumnDelegate(self))
            self.setModel(self.ruleModel)
            if Debug.modelTest:
                self.ruleModelTest = ModelTest(self.ruleModel, self)
            self.show()

    def selectionChanged(self, selected:'QItemSelection', unused_deselected:Optional['QItemSelection']=None) ->None:
        """update editing buttons"""
        assert self.ruleModel
        enableCopy = enableRemove = enableCompare = False
        if selected.indexes():
            item = cast(RuleTreeItem, selected.indexes()[0].internalPointer())
            isPredefined = isinstance(item.ruleset(), PredefinedRuleset)
            if isinstance(item, RulesetItem):
                enableCompare = True
                enableCopy = sum(
                    x.hash == item.ruleset(
                    ).hash for x in self.ruleModel.rulesets) == 1
                enableRemove = not isPredefined
        if self.btnCopy:
            self.btnCopy.setEnabled(enableCopy)
        if self.btnRemove:
            self.btnRemove.setEnabled(enableRemove)
        if self.btnCompare:
            self.btnCompare.setEnabled(enableCompare)

    def showEvent(self, unusedEvent:Optional['QShowEvent']) ->None:
        """reload the models when the view comes into sight"""
        # default: make sure the name column is wide enough
        assert self.ruleModel
        if self.ruleModel.canFetchMore():
            # we want to load all before adjusting column width
            self.ruleModel.fetchMore()
        if header := self.header():
            header.setStretchLastSection(False)
            header.setMinimumSectionSize(-1)
            for col in range(1, header.count()):
                header.setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
            header.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
            for col in range(header.count()):
                self.resizeColumnToContents(col)

    def selectedRow(self) ->Optional[QModelIndex]:
        """return the currently selected row index (with column 0)"""
        if model := self.selectionModel():
            rows = model.selectedRows()
            if len(rows) == 1:
                return rows[0]
        return None

    def copyRow(self) ->None:
        """copy a ruleset"""
        row = self.selectedRow()
        if row:
            item = row.internalPointer()
            assert isinstance(item, RulesetItem)
            ruleset = item.raw.copyTemplate()
            cast(RuleModel, self.model()).appendRuleset(ruleset)
            self.rulesets.append(ruleset)
            if model := self.selectionModel():
                self.selectionChanged(model.selection())

    def removeRow(self) ->None:
        """removes a ruleset or a rule"""
        row = self.selectedRow()
        if row:
            item = cast(RuleTreeItem, row.internalPointer())
            assert not isinstance(item.ruleset(), PredefinedRuleset)
            assert isinstance(item, RulesetItem)
            ruleset = item.ruleset()
            if model := self.model():
                model.removeRow(row.row(), parent=row.parent())
            self.rulesets.remove(ruleset)
            if sel_model := self.selectionModel():
                self.selectionChanged(sel_model.selection())

    def compareRow(self) ->None:
        """shows the difference between two rulesets"""
        if row0 := self.selectedRow():
            ruleset = cast(RuleTreeItem, row0.internalPointer()).raw
            assert isinstance(ruleset, Ruleset)
            differ = RulesetDiffer([ruleset], self.rulesets)
            differ.show()
            self.differs.append(differ)


class RulesetSelector(QWidget):

    """presents all available rulesets with previews"""

    def __init__(self, parent:Optional[QWidget]=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setupUi()

    def setupUi(self) ->None:
        """layout the window"""
        decorateWindow(self, i18nc("@title:window", "Customize rulesets"))
        self.setObjectName('Rulesets')
        hlayout = QHBoxLayout(self)
        v1layout = QVBoxLayout()
        self.v1widget = QWidget()
        v1layout = QVBoxLayout(self.v1widget)
        v2layout = QVBoxLayout()
        hlayout.addWidget(self.v1widget)
        hlayout.addLayout(v2layout)
        widgets:List[Union[QWidget, 'QLayout']] = [self.v1widget, hlayout, v1layout, v2layout]
        for widget in widgets:
            widget.setContentsMargins(0, 0, 0, 0)
        hlayout.setStretchFactor(self.v1widget, 10)
        self.btnCopy = QPushButton()
        self.btnRemove = QPushButton()
        self.btnCompare = QPushButton()
        self.btnClose = QPushButton()
        self.rulesetView = RuleTreeView(
            i18ncE('kajongg',
                   'Rule'),
            self.btnCopy,
            self.btnRemove,
            self.btnCompare)
        v1layout.addWidget(self.rulesetView)
        self.rulesetView.setWordWrap(True)
        self.rulesetView.setMouseTracking(True)
        spacerItem = QSpacerItem(
            20,
            20,
            QSizePolicy.Policy.Minimum,
            QSizePolicy.Policy.Expanding)
        v2layout.addWidget(self.btnCopy)
        v2layout.addWidget(self.btnRemove)
        v2layout.addWidget(self.btnCompare)
        self.btnCopy.clicked.connect(self.rulesetView.copyRow)
        self.btnRemove.clicked.connect(self.rulesetView.removeRow)
        self.btnCompare.clicked.connect(self.rulesetView.compareRow)
        self.btnClose.clicked.connect(self.hide)
        v2layout.addItem(spacerItem)
        v2layout.addWidget(self.btnClose)
        self.retranslateUi()
        StateSaver(self)
        self.show()

    def sizeHint(self) ->QSize:
        """we never want a horizontal scrollbar for player names,
        we always want to see them in full"""
        result = QWidget.sizeHint(self)
        available = KApplication.desktopSize()
        height = max(result.height(), available.height() * 2 // 3)
        width = max(result.width(), available.width() // 2)
        return QSize(width, height)

    def minimumSizeHint(self) ->QSize:
        """we never want a horizontal scrollbar for player names,
        we always want to see them in full"""
        return self.sizeHint()

    def showEvent(self, unusedEvent:Optional['QShowEvent']) ->None:
        """reload the rulesets"""
        self.refresh()

    def refresh(self) ->None:
        """retranslate and reload rulesets"""
        self.retranslateUi()
        self.rulesetView.rulesets = Ruleset.availableRulesets()

    def hideEvent(self, event:Optional['QHideEvent']) ->None:
        """close all differ dialogs"""
        marked:List[RulesetDiffer] = []
        differs = self.rulesetView.differs
        for differ in differs:
            differ.hide()
            marked.append(differ)
        if event:
            super().hideEvent(event)
        for _ in marked:
            del differs[_]

    def retranslateUi(self) ->None:
        """translate to current language"""
        self.btnCopy.setText(i18n("C&opy"))
        self.btnCompare.setText(i18nc('Kajongg ruleset comparer', 'Co&mpare'))
        self.btnRemove.setText(i18n('&Remove'))
        self.btnClose.setText(i18n('&Close'))



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
 - Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

import os
import traceback
import datetime
import random
import sqlite3
from collections import namedtuple
from typing import List, Tuple, Union, Any, Optional, Literal, Generator

from mi18n import i18n, i18ncE
from util import Duration
from log import logInfo, logWarning, logException, logError, logDebug
from common import Options, Internal, Debug, appdataDir, ReprMixin

class QueryException(Exception):

    """as the name says"""

    def __init__(self, msg:str) ->None:
        super().__init__(msg)


class DBCursor(sqlite3.Cursor, ReprMixin):

    """logging wrapper"""

    def __init__(self, dbHandle:'DBHandle') ->None:
        super().__init__(dbHandle)
        self.statement:str
        self.parameters:Optional[Tuple[Union[str, int], ...]] = None

    def execute(self, statement:str, parameters:Optional[Tuple[Union[str, int], ...]]=None, # type:ignore[override]
                silent:bool=False, failSilent:bool=False, mayFail:bool=False) ->None:
        """logging wrapper, returning all selected data"""
        self.statement = statement
        self.parameters = parameters
        if not silent:
            logDebug(repr(self))
        try:
            with Duration(f'{self!r}', 60.0 if Debug.neutral else 3.0):
                if isinstance(parameters, list):
                    sqlite3.Cursor.executemany(
                        self, statement, parameters)
                elif parameters:
                    super().execute(statement, parameters)
                else:
                    super().execute(statement)
        except sqlite3.Error as exc:
            msg = f"{self!r}: ERROR {exc}"
            if mayFail:
                if not failSilent and Debug.sql:
                    logDebug(msg)
            else:
                if not failSilent:
                    logError(msg)
                raise

    def __str__(self) ->str:
        """the statement"""
        if self.parameters is not None:
            return f"{self.connection}:{self.statement} [{self.parameters}]"
        return f"{self.connection}:{self.statement}"


class DBHandle(sqlite3.Connection, ReprMixin):

    """a handle with our preferred configuration"""

    def __init__(self, path: str) ->None:
        assert Internal.db is None, id(self)
        Internal.db = self
        self.inTransaction:Optional[datetime.datetime] = None
        self.path = path
        self.identifier = None
        try:
            super().__init__(self.path, timeout=10.0, detect_types=sqlite3.PARSE_DECLTYPES)
        except sqlite3.Error as exc:
            if hasattr(exc, 'message'):
                msg = exc.message
            elif hasattr(exc, 'args'):
                msg = ' '.join(exc.args)
            else:
                msg = ''
            logException(f'opening {self.path}: {msg}')
        if self.hasTable('general'):
            cursor = self.cursor()
            cursor.execute('select ident from general')
            self.identifier = cursor.fetchone()[0]
        if Debug.sql:
            logDebug(f'{self} opened with identifier {self.identifier}')

    def __enter__(self) ->'DBHandle':
        self.inTransaction = datetime.datetime.now()
        if Debug.sql:
            logDebug('starting transaction')
        return super().__enter__()

    def __exit__(self, *args:Any) ->Literal[False]:
        super().__exit__(*args)
        if Debug.sql:
            logDebug('finished transaction')
        self.inTransaction = None
        return False

    def __str__(self):
        return f'{self.dbPath()}'

    @staticmethod
    def dbPath() ->str:
        """
        The path for the data base.

        @return: The full path for kajonggserver.db or kajongg.db.
        @rtype: C{str}
        """
        name = 'kajonggserver' if Internal.isServer else 'kajongg'
        name += '3.db'
        return Options.dbPath if Options.dbPath else os.path.join(appdataDir(), name)

    @property
    def debug_name(self) ->str:
        """get name for log messages. Readonly."""
        stack = [x[2] for x in traceback.extract_stack()]
        name = stack[-3]
        if name in ('__exit__', '__init__'):
            name = stack[-4]
        return f'{name} on {self!r}'

    def commit(self, silent:bool=False) ->None:
        """commit and log it"""
        try:
            sqlite3.Connection.commit(self)
        except sqlite3.Error as exc:
            if not silent:
                logWarning(
                    f'{self!r} cannot commit: {exc} :')

    def rollback(self, silent:bool=False) ->None:
        """rollback and log it"""
        try:
            sqlite3.Connection.rollback(self)
            if not silent and Debug.sql:
                logDebug(f'{self!r} rollback')
        except sqlite3.Error as exc:
            logWarning(f'{self!r} cannot rollback: {exc}')

    def close(self, silent:bool=False) ->None:
        """just for logging"""
        if not silent and (Debug.sql or Debug.quit):
            if self is Internal.db:
                logDebug(f'{self!r} closing Internal.db')
            else:
                logDebug(f'{self!r} closing')
        if self is Internal.db:
            Internal.db = None
        try:
            self.commit(silent=True)
        except sqlite3.Error:
            self.rollback()
        try:
            sqlite3.Connection.close(self)
        except sqlite3.Error as exc:
            logDebug(exc)

    @staticmethod
    def hasTable(table:str) ->bool:
        """does the table contain table?"""
        return len(Query(f'SELECT name FROM sqlite_master WHERE type="table" AND name="{table}"').records) > 0

    def tableHasField(self, table:str, field:str) ->bool:
        """does the table contain a column named field?"""
        cursor = self.cursor()
        cursor.execute(f'select * from {table}')
        return any(x[0] == field for x in cursor.description)


class Query(ReprMixin):

    """a wrapper arout python sqlite3, adding logging and some exception handling.
    For selecting queries we fill a list with ALL records which is never much.
    Every record is a list of all fields. q.records[0][1] is record 0, field 1.
    """

    localServerName = i18ncE(
        'kajongg name for local game server',
        'Local Game')

    def __init__(self, statement:str,
                 args:Union[None,Tuple[Union[str, int, float], ...],List[List[Union[str,int,float]]]]=None,
                 silent:bool=False, mayFail:bool=False, failSilent:bool=False,
                 fields:Optional[str]=None) ->None:
        """we take one sql statement.
        Do prepared queries by passing the parameters in args.
        If args is a list of lists, execute the prepared query for every sublist.
        Use Internal.db for db access.
        Else if the default dbHandle (Internal.db) is defined, use it."""
        silent |= not Debug.sql
        self.msg = None
        self.records:List[List[Any]] = []
        self.statement = statement
        self.args = args
        self.fields = fields  # a field name might be "lastname as playername"
        self.tuplefields:Optional[List[str]] = None
        if fields:
            self.tuplefields = list(self.__tuple_fieldnames())
            self.statement = self.statement.format(fields=fields)

        if Internal.db:
            self.cursor = Internal.db.cursor(
                DBCursor)
            self.cursor.execute(
                self.statement,
                args,
                silent=silent,
                mayFail=mayFail,
                failSilent=failSilent)
            self.records = list(self.cursor.fetchall())
            if not Internal.db.inTransaction:
                Internal.db.commit()
        else:
            # may happen at shutdown
            logDebug(f'Internal.db is None for {self}')
            self.cursor = None
            self.records = []
        if self.records and Debug.sql:
            logDebug(f'result set:{self.records}')

    def __tupleName(self) ->str:
        """the class name of the returned tuples.
           For now:

           - use first table name from statement.
           - If none found: use 'Fields'

           Query might get an arg tupleName which would override this
        """
        lower = self.statement.lower()
        parts = lower.split(' from ')
        if len(parts) > 1:
            return 'Query_' + parts[1].split()[0].strip().capitalize()
        return 'Query_Fields'

    def __tuple_fieldnames(self) ->Generator[str, None, None]:
        """translates 'fieldname as myname' into 'myname' """
        if not isinstance(self.fields, str):
            raise ValueError(f'must be str:{self.fields}')
        for _ in self.fields.split(','):
            words = [x.strip() for x in _.strip().split(' ')]
            if len(words) == 1:
                yield words[0]
            elif len(words) == 3 and words[1] == 'as':
                yield words[2]
            else:
                raise ValueError(f'cannot parse {_} out of {self.fields}, words={words}')

    def tuples(self) -> List[Any]:
        """named tuples for query records"""
        tupleclass = namedtuple(self.__tupleName(), self.tuplefields)  # type: ignore
        return [tupleclass._make(x) for x in self.records]

    def tuple(self) -> Any:
        """Valid only for queries returning exactly one record"""
        if not self.fields:
            raise ValueError
        if len(self.records) != 1:
            raise ValueError(f'{self!r} did not return exactly 1 record but {len(self.records)}')
        tupleclass = namedtuple('Fields', self.tuplefields)  # type: ignore
        return tupleclass._make(self.records[0])

    def record(self) -> Any:
        """Valid only for queries returning exactly one record"""
        if len(self.records) != 1:
            raise ValueError(f'{self!r} did not return exactly 1 record but {len(self.records)}')
        return self.records[0]

    def __str__(self) ->str:
        return f"{Internal.db!r}:{self.statement} {'args=' + ','.join(str(x) for x in self.args) if self.args else ''}"

    def rowcount(self) ->int:
        """how many rows were affected?"""
        return self.cursor.rowcount if self.cursor else 0

    def map(self, namedTuple) ->Optional[List[Any]]:  # FIXME: remove again
        """A list of namedtuple"""
        if self.records is None:
            return None
        return [namedTuple._make(x) for x in self.records]

    def column(self, idx:int) -> List[Any]:
        """A list with only that column"""
        return [x[idx] for x in self.records]

def initDb() ->bool:
    """open the db, create or update it if needed.
    sets Internal.db."""
    PrepareDB(DBHandle.dbPath())  # create or upgrade
    DBHandle(DBHandle.dbPath())
    return True


class PrepareDB:

    """create or upgrade DB if needed"""
    schema = {}
    schema['player'] = """
        id INTEGER PRIMARY KEY,
        name TEXT unique"""
    schema['game'] = """
            id integer primary key,
            seed text,
            autoplay integer default 0,
            starttime text default current_timestamp,
            endtime text,
            ruleset integer references ruleset(id),
            p0 integer constraint fk_p0 references player(id),
            p1 integer constraint fk_p1 references player(id),
            p2 integer constraint fk_p2 references player(id),
            p3 integer constraint fk_p3 references player(id)"""
    schema['score'] = """
            game integer constraint fk_game references game(id),
            hand integer,
            data text,
            manualrules text,
            rotated integer,
            notrotated integer,
            player integer constraint fk_player references player(id),
            scoretime text,
            won integer,
            penalty integer default 0,
            prevailing wind,
            wind wind,
            points integer,
            payments integer,
            balance integer"""
    schema['ruleset'] = """
            id integer primary key,
            name text,
            hash text,
            description text"""
    schema['rule'] = """
            ruleset integer,
            list integer,
            position integer,
            name text,
            definition text,
            points text,
            doubles text,
            limits text,
            parameter text,
            primary key(ruleset,list,position)"""
    schema['server'] = """
                url text,
                lastname text,
                lasttime text,
                lastruleset integer,
                primary key(url)"""
    schema['passwords'] = """
                url text,
                player integer,
                password text"""
    schema['general'] = """
                ident text,
                schemaversion text"""

    def __init__(self, path:str) ->None:
        self.path = path
        if not os.path.exists(path):
            self.__create()
        else:
            self.__upgrade()

    def __create(self) ->None:
        """create a brand new kajongg database"""
        tmpPath = f'{self.path}.new.{int(os.getpid())}'
        Internal.db = DBHandle(tmpPath)
        try:
            with Internal.db:
                self.createTables()
                self.__generateDbIdent()
                Query(
                    'UPDATE general SET schemaversion=?', (Internal.defaultPort,))
        finally:
            Internal.db.close(silent=True)
        if os.path.exists(self.path):
            # somebody was faster
            os.remove(tmpPath)
        else:
            os.rename(tmpPath, self.path)

    @staticmethod
    def __currentVersion() ->str:
        """
        Get current version of DB schema as a comparable string.

        @returns: The current version from the database.
        @rtype: C{str}
        """
        if Internal.db.tableHasField('general', 'schemaversion'):
            return Query('select schemaversion from general').records[0][0]
        return '1.1.1'

    def __upgrade(self) ->None:
        """upgrade the structure of an existing kajongg database"""
        try:
            Internal.db = DBHandle(self.path)
            allVersions = list(['4.13.0', '8300', '8301', '8302', '8303'])
            assert allVersions[-1] == str(Internal.defaultPort), f'{allVersions[-1]} != {str(Internal.defaultPort)}'
            # skip versions before current db versions:
            currentVersion = self.__currentVersion()
            while allVersions and allVersions[0] <= currentVersion:
                allVersions = allVersions[1:]
            for version in allVersions:
                currentVersion = self.__currentVersion()
                with Internal.db:  # transaction
                    updateMethodName = f"updateToVersion{version.replace('.', '_')}"
                    if hasattr(self, updateMethodName):
                        getattr(self, updateMethodName)()
                    Query('UPDATE general SET schemaversion=?', (version,))
                logInfo(i18n('Database %1 updated from schema %2 to %3',
                             Internal.db.path, currentVersion, version), showDialog=True)
        except sqlite3.Error as exc:
            logException(f'opening {self.path}: {exc}')
        finally:
            Internal.db.close(silent=True)

    def updateToVersion8302(self) ->None:
        """seed now is 0 instead of Null"""
        Query('UPDATE game SET seed=0 WHERE seed IS NULL', silent=True)

    def __needs_type_wind(self) ->bool:
        """helper for updating to 8303"""
        rows = Query("PRAGMA table_info('score')").records
        for row in rows:
            name = row[1]
            typ = row[2]
            if name == 'wind' and typ != 'wind':
                logInfo(f'score.wind has not yet type wind: {typ}')
                return True
            if name == 'prevailing' and typ != 'wind':
                logInfo(f'score.prevailing has not yet type wind: {typ}')
                return True
        return False

    def updateToVersion8303(self) ->None:
        """new type 'wind'"""
        if self.__needs_type_wind():
            with Internal.db:
                Query('ALTER TABLE score add prevailingw wind', mayFail=True)
                Query('ALTER TABLE score add windw wind', mayFail=True)
                Query('UPDATE score set prevailingw = prevailing, windw=wind')
                Query('ALTER TABLE score drop prevailing')
                Query('ALTER TABLE score drop wind')
                Query('ALTER TABLE score rename windw to wind')
                Query('ALTER TABLE score rename prevailingw to prevailing')

    @classmethod
    def sqlForCreateTable(cls, table:str) ->str:
        """the SQL command for creating 'table'"""
        return f"create table {table}({cls.schema[table]})"

    @classmethod
    def createTable(cls, table:str) ->None:
        """create a single table using the predefined schema"""
        if not Internal.db.hasTable(table):
            Query(cls.sqlForCreateTable(table))

    @classmethod
    def createTables(cls) ->None:
        """creates empty tables"""
        for table in ['player', 'game', 'score', 'ruleset', 'rule', 'general']:
            cls.createTable(table)
        cls.createIndex('idxgame', 'score(game)')
        # this makes finding suspended games much faster in the presence
        # of many test games (with autoplay=1)
        cls.createIndex('idxautoplay', 'game(autoplay)')

        if Internal.isServer:
            Query('ALTER TABLE player add password text')
        else:
            cls.createTable('passwords')
            cls.createTable('server')

    @staticmethod
    def createIndex(name:str, cmd:str) ->None:
        """only try to create it if it does not yet exist. Do not use create if not exists because
        we want debug output only if we really create the index"""
        if not Query(
                "select 1 from sqlite_master where type='index' and name=?", (
                    name,),
                silent=True).records:
            Query(f"create index {name} on {cmd}")

    @staticmethod
    def __generateDbIdent() ->None:
        """make sure the database has a unique ident and get it"""
        records = Query('select ident from general').records
        assert len(records) < 2
        if not records:
            dbIdent = str(random.randrange(100000000000))
            Query("INSERT INTO general(ident) values(?)", (dbIdent,))
            if Debug.sql:
                logDebug(
                    f'generated new dbIdent {dbIdent} for {Internal.db.path}')



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Sequence, Tuple, Union, Optional, Any, Dict, List, cast

import weakref
from itertools import chain

from message import Message
from common import IntDict, Debug, ReprMixin
from tile import Tile

if TYPE_CHECKING:
    from player import PlayingPlayer
    from tile import Tiles, TileTuple, MeldList, Meld
    from hand import Hand

class AIDefaultAI:

    """all AI code should go in here"""

    groupPrefs = dict(zip(Tile.colors + Tile.honors, (0, 0, 0, 4, 7)))

    # we could solve this by moving those filters into DiscardCandidates
    # but that would make it more complicated to define alternative AIs

    def __init__(self, player:'PlayingPlayer') ->None:
        self._player = weakref.ref(player)

    @property
    def player(self) ->'PlayingPlayer':
        """hide weakref"""
        result = self._player()
        assert result
        return result

    def name(self) ->str:
        """return our name"""
        return self.__class__.__name__[2:]

    @staticmethod
    def weighSameColors(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """weigh tiles of same group against each other"""
        for candidate in candidates:
            if candidate.group in Tile.colors:
                if candidate.prev.occurrence:
                    candidate.prev.keep += 1.001
                    candidate.keep += 1.002
                    if candidate.next.occurrence:
                        candidate.prev.keep += 2.001
                        candidate.next.keep += 2.003
                if candidate.next.occurrence:
                    candidate.next.keep += 1.003
                    candidate.keep += 1.002
                elif candidate.next2.occurrence:
                    candidate.keep += 0.502
                    candidate.next2.keep += 0.503
        return candidates

    def selectDiscard(self, hand:'Hand') ->Tile:
        # disable warning about too many branches
        """return exactly one tile for discard.
        Much of this is just trial and success - trying to get as much AI
        as possible with limited computing resources, it stands on
        no theoretical basis"""
        candidates = DiscardCandidates(self.player, hand)
        result = self.weighDiscardCandidates(candidates).best()
        candidates.unlink()
        return result

    def weighDiscardCandidates(self, candidates:'DiscardCandidates') ->'DiscardCandidates':
        """the standard"""
        game = self.player.game
        assert game
        weighRules = game.ruleset.filterRules('weigh')
        for _ in [self.weighBasics, self.weighSameColors,
                         self.weighSpecialGames, self.weighCallingHand,
                         self.weighOriginalCall,
                         self.alternativeFilter] + weighRules:
            if _ in weighRules:
                filterName = _.__class__.__name__
                aiFilter = _.weigh
            else:
                filterName = _.__name__
                aiFilter = _
            if Debug.robotAI:
                prevWeights = ((x.tile, x.keep) for x in candidates)
                candidates = aiFilter(self, candidates)
                newWeights = ((x.tile, x.keep) for x in candidates)
                for oldW, newW in zip(prevWeights, newWeights):
                    if oldW != newW:
                        game.debug(f'{filterName}: {oldW[0]}: {oldW[1]:.3f}->{newW[1]:.3f}')
            else:
                candidates = aiFilter(self, candidates)
        return candidates

    @staticmethod
    def alternativeFilter(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """if the alternative AI only adds tests without changing
        default filters, you can override this one to minimize
        the source size of the alternative AI"""
        return candidates

    @staticmethod
    def weighBasics(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """basic things"""
        # pylint: disable=too-many-branches
        # too many branches
        for candidate in candidates:
            keep = candidate.keep
            tile = candidate.tile
            value = tile.value
            if candidate.dangerous:
                keep += 1000
            if candidate.occurrence >= 3:
                keep += 10.04
            elif candidate.occurrence == 2:
                keep += 5.08
            keep += aiInstance.groupPrefs[tile.group]
            if tile.isWind:
                if value == candidates.hand.ownWind:
                    keep += 1.01
                if value == candidates.hand.roundWind:
                    keep += 1.02
            if tile.isTerminal:
                keep += 2.16
            if candidate.maxPossible == 1:
                if tile.isHonor:
                    keep -= 8.32
                    # not too much, other players might profit from this tile
                else:
                    if not candidate.next.maxPossible:
                        if not candidate.prev.maxPossible or not candidate.prev2.maxPossible:
                            keep -= 100
                    if not candidate.prev.maxPossible:
                        if not candidate.next.maxPossible or not candidate.next2.maxPossible:
                            keep -= 100
            if candidate.available == 1 and candidate.occurrence == 1:
                if tile.isHonor:
                    keep -= 3.64
                else:
                    if not candidate.next.maxPossible:
                        if not candidate.prev.maxPossible or not candidate.prev2.maxPossible:
                            keep -= 3.64
                    if not candidate.prev.maxPossible:
                        if not candidate.next.maxPossible or not candidate.next2.maxPossible:
                            keep -= 3.64
            candidate.keep = keep
        return candidates

    @staticmethod
    def weighSpecialGames(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """like color game, many dragons, many winds"""
        # pylint: disable=too-many-nested-blocks
        for candidate in candidates:
            tile = candidate.tile
            groupCount = candidates.groupCounts[tile.group]
            if tile.isWind:
                if groupCount > 8:
                    candidate.keep += 10.153
            elif tile.isDragon:
                if groupCount > 7:
                    candidate.keep += 15.157
            else:
                # count tiles with a different group:
                if groupCount == 1:
                    candidate.keep -= 2.013
                else:
                    otherGC = sum(candidates.groupCounts[x]
                                  for x in Tile.colors if x != tile.group)
                    if otherGC:
                        if groupCount > 8 or otherGC < 5:
                            # do not go for color game if we already declared
                            # something in another group:
                            if not any(candidates.declaredGroupCounts[x] for x in Tile.colors if x != tile.group):
                                candidate.keep += 20 // otherGC
        return candidates

    @staticmethod
    def weighOriginalCall(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """if we declared Original Call, respect it"""
        myself = aiInstance.player
        game = myself.game
        assert game
        if myself.originalCall and myself.mayWin:
            last = myself.lastTile
            if Debug.originalCall:
                game.debug(f'weighOriginalCall: lastTile={last}, candidates={[str(x) for x in candidates]}')
            for candidate in candidates:
                if last and last.exposed and candidate.tile:
                    assert myself.originalCallingHand
                    winningTiles = myself.originalCallingHand.chancesToWin()
                    if Debug.originalCall:
                        game.debug(f'weighOriginalCall: winningTiles={winningTiles} '
                                   f'for {str(myself.originalCallingHand)}')
                        game.debug(f'weighOriginalCall respects originalCall: '
                                   f'{candidate.tile} with {int(-99 * len(winningTiles))}')
                    candidate.keep -= 99 * len(winningTiles)
        return candidates

    @staticmethod
    def weighCallingHand(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        """if we can get a calling hand, prefer that"""
        assert aiInstance.player.game
        for candidate in candidates:
            newHand = candidates.hand - candidate.tile.concealed
            winningTiles = newHand.chancesToWin()
            if winningTiles:
                for winnerTile in sorted(set(winningTiles)):
                    winnerHand = newHand + winnerTile.concealed
                    if Debug.robotAI:
                        aiInstance.player.game.debug(f"weighCallingHand {newHand} cand {candidate} "
                                                     f"winnerTile {winnerTile} "
                                                     f"winnerHand {winnerHand}: "
                                                     f"{'     '.join(winnerHand.explain())}")
                    keep = winnerHand.total() / 10.017
                    candidate.keep -= keep
                    if Debug.robotAI:
                        aiInstance.player.game.debug(
                            f'weighCallingHand {newHand} winnerTile {winnerTile}: '
                            f'discardCandidate {candidate} keep -= {keep:.4f}')
                # more weight if we have several chances to win
                candidate.keep -= float(len(winningTiles)) / len(
                    set(winningTiles)) * 5.031
                if Debug.robotAI:
                    assert aiInstance.player.game
                    aiInstance.player.game.debug(f'weighCallingHand {newHand} for '
                                                 f'{candidates.hand} winningTiles:{winningTiles}')
        return candidates

    def selectAnswer(self, answers:Sequence[Message]) ->Tuple[Message, Union[Tile, 'Meld', None]]:
        """this is where the robot AI should go.
        Returns answer and one parameter"""
        # disable warning about too many branches
        assert self.player.game
        answer:Message = answers[0]
        parameter:Any = None
        tryAnswers = (
            x for x in [Message.MahJongg, Message.OriginalCall, Message.Kong,
                        Message.Pung, Message.Chow, Message.Discard] if x in answers)
        hand = self.player.hand
        claimness = IntDict()
        discard = self.player.game.lastDiscard
        if discard:
            for rule in self.player.game.ruleset.filterRules('claimness'):
                claimness += rule.claimness(hand, discard)  # type:ignore[attr-defined]
                if Debug.robotAI:
                    hand.debug(
                        f'{rule.name}: claimness in selectAnswer:{claimness}')
        for tryAnswer in tryAnswers:
            parameter = self.player.sayable[tryAnswer]
            if not parameter:
                continue
            if claimness[tryAnswer] < 0:
                continue
            if tryAnswer in [Message.Discard, Message.OriginalCall]:
                parameter = self.selectDiscard(hand)
            elif tryAnswer == Message.Pung and self.player.maybeDangerous(tryAnswer):
                continue
            elif tryAnswer == Message.Chow:
                parameter = self.selectChow(cast('MeldList', parameter))
            elif tryAnswer == Message.Kong:
                parameter = self.selectKong(cast('MeldList', parameter))
            if parameter:
                answer = tryAnswer
                break
        return answer, parameter

    def selectChow(self, chows:'MeldList') ->Optional['Meld']:
        """selects a chow to be completed. Add more AI here."""
        for chow in chows:
            # a robot should never play dangerous
            if not self.player.mustPlayDangerous(chow):
                if not self.player.hasConcealedTiles(chow):
                    # do not dissolve an existing chow
                    belongsToPair = False
                    for tile in chow:
                        if self.player.concealedTiles.count(tile) == 2:
                            belongsToPair = True
                            break
                    if not belongsToPair:
                        return chow
        return None

    def selectKong(self, kongs:'MeldList') ->Optional['Meld']:
        """selects a kong to be declared. Having more than one undeclared kong is quite improbable"""
        for kong in kongs:
            if not self.player.mustPlayDangerous(kong):
                return kong
        return None

    def handValue(self, hand:'Hand') ->int:
        """compute the value of a hand.
        This is not just its current score but also
        what possibilities to evolve it has. E.g. if
        only one tile is concealed and 3 of it are already
        visible, chances for MJ are 0.
        This will become the central part of AI -
        moves will be done which optimize the hand value.
        For now it is only used by Hand.__split but not
        by the actual discarding code"""
        return hand.total()

# the rest is not yet used: __split only wants something nice for
# display but is not relevant for the real decision making
#        if hand is None:
#            hand = self.player.hand
#        result = 0
#        if hand.won:
#            return 1000 + hand.total()
#        result = hand.total()
#        if hand.callingHands:
#            result += 500 + len(hand.callingHands) * 20
#        for meld in hand.declaredMelds:
#            if not meld.isChow:
#                result += 40
#        garbage = []
#        for meld in (x for x in hand.melds if not x.isDeclared):
#            assert len(meld) < 4, hand
#            if meld.isPung:
#                result += 50
#            elif meld.isChow:
#                result += 30
#            elif meld.isPair:
#                result += 5
#            else:
#                garbage.append(meld)
#        return result


class TileAI(ReprMixin):

    """holds a few AI related tile properties"""
    # pylint: disable=too-many-instance-attributes
    # we do want that many instance attributes

    def __init__(self, candidates:'DiscardCandidates', tile:Tile) ->None:
        assert candidates.player
        assert candidates.player.game
        self.tile = tile
        self.group, self.value = tile.group, tile.value
        if tile.isReal:
            self.occurrence = candidates.hiddenTiles.count(tile)
            self.available = candidates.player.tileAvailable(
                tile, candidates.hand)
            self.maxPossible = self.available + self.occurrence
            self.dangerous = bool(
                candidates.player.game.dangerousFor(
                    candidates.player,
                    tile))
        else:
            # value might be -1, 0, 10, 11 for suits
            self.occurrence = 0
            self.available = 0
            self.maxPossible = 0
            self.dangerous = False
        self.keep = 0.0
        self.prev = None
        self.next = None
        self.prev2 = None
        self.next2 = None

    def __lt__(self, other:object) ->bool:
        """for sorting"""
        return self.tile < cast('TileAI', other).tile

    def __str__(self) ->str:
        dang = f' dang:{int(self.dangerous)}' if self.dangerous else ''
        return f'{self.tile}:={self.keep:.4f}{dang}'


class DiscardCandidates(list):

    """a list of TileAI objects. This class should only hold
    AI neutral methods"""

    def __init__(self, player:'PlayingPlayer', hand:'Hand') ->None:
        list.__init__(self)
        self._player = weakref.ref(player)
        self._hand = weakref.ref(hand)
        if Debug.robotAI:
            assert player.game
            player.game.debug(f'DiscardCandidates for hand {hand} are {hand.tilesInHand}')
        self.hiddenTiles:List[Tile] = [x.exposed for x in hand.tilesInHand]
        self.groupCounts:Dict[str, int] = IntDict()
                                   # counts for tile groups (sbcdw), exposed
                                   # and concealed
        for tile in self.hiddenTiles:
            self.groupCounts[tile.group] += 1
        self.declaredGroupCounts:Dict[str, int] = IntDict()
        for tile in chain(*hand.declaredMelds):
            self.groupCounts[tile.lowerGroup] += 1
            self.declaredGroupCounts[tile.lowerGroup] += 1
        self.extend(TileAI(self, x) for x in sorted(set(self.hiddenTiles)))
        self.link()

    @property
    def player(self) ->'PlayingPlayer':
        """hide weakref"""
        result = self._player()
        assert result
        return result

    @property
    def hand(self) ->'Hand':
        """hide weakref"""
        result = self._hand()
        assert result
        return result

    def link(self) ->None:
        """define values for candidate.prev and candidate.next"""
        prev = prev2 = None
        for this in self:
            if this.group in Tile.colors:
                thisValue = this.value
                if prev and prev.group == this.group:
                    if prev.value + 1 == thisValue:
                        prev.next = this
                        this.prev = prev
                    if prev.value + 2 == thisValue:
                        prev.next2 = this
                        this.prev2 = prev
                if prev2 and prev2.group == this.group and prev2.value + 2 == thisValue:
                    prev2.next2 = this
                    this.prev2 = prev2
            prev2 = prev
            prev = this
        for this in self:
            if this.group in Tile.colors:
                # we want every tile to have prev/prev2/next/next2
                # the names do not matter, just occurrence, available etc
                thisValue = this.value
                if not this.prev:
                    this.prev = TileAI(self, this.tile.prevForChow)
                if not this.prev2:
                    this.prev2 = TileAI(self, this.prev.tile.prevForChow)
                if not this.next:
                    this.next = TileAI(self, this.tile.nextForChow)
                if not this.next2:
                    this.next2 = TileAI(self, this.next.tile.nextForChow)

    def unlink(self) ->None:
        """remove links between elements. This helps garbage collection."""
        for this in self:
            this.prev = None
            this.next = None
            this.prev2 = None
            this.next2 = None

    def best(self) ->Tile:
        """return the candidate with the lowest value"""
        lowest = min(x.keep for x in self)
        candidates = sorted(x for x in self if x.keep == lowest)
        assert self.player.game
        result = self.player.game.randomGenerator.choice(
            candidates).tile.concealed
        if Debug.robotAI:
            self.player.game.debug(
                f"{self.player}: discards {result} out of {' '.join(str(x) for x in self)}")
        return result



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime
import weakref
import sys
from collections import namedtuple

from typing import List, Optional, Tuple, TYPE_CHECKING, Union, Dict, Iterable, Generator, Any, Set, cast

from twisted.internet.defer import succeed
from util import gitHead
from kajcsv import CsvRow
from rand import CountingRandom
from log import logError, logException, logDebug, i18n
from common import Internal, IntDict, Debug, Options
from common import Speeds
from wind import Wind, East, South, West, North
from query import Query
from rule import Ruleset, UsedRule
from tile import Tile, elements
from tilesource import TileSource
from sound import Voice
from wall import Wall
from player import Players, Player, PlayingPlayer
from animation import animateAndDo, AnimationSpeed, ParallelAnimationGroup
from point import Point, PointRange

if sys.platform != 'win32':
    import resource

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred
    from move import Move
    from client import Client
    from uiwall import UIWall
    from message import Message
    from handboard import PlayingHandBoard
    from servertable import ServerGame


class Game:

    """the game without GUI"""
    # pylint: disable=too-many-instance-attributes
    playerClass = Player
    wallClass = Wall

    def __init__(self, names:List[Tuple['Wind', str]], ruleset:Ruleset, gameid:Optional[int]=None,
                 wantedGame:Optional[str]=None, client:Optional['Client']=None):
        """a new game instance. May be shown on a field, comes from database
        if gameid is set.

        Game.lastDiscard is the tile last discarded by any player. It is
        reset to None when a player gets a tile from the living end of the
        wall or after he claimed a discard.
        """
        assert self.__class__ != Game, 'Do not directly instantiate Game'
        for wind, name in names:
            assert isinstance(wind, Wind), 'Game.__init__ expects Wind objects'
            assert isinstance(name, str), f'Game.__init__: name must be string and not {type(name)}'
        self.players = Players()
        # if we fail later on in init, at least we can still close the program
        self.myself:Player
        self.prevActivePlayer:Optional[Player]
        # the player using this client instance for talking to the server
        self.__shouldSave = False
        self._client = None  # FIXME: add comment
        self.client = client  # FIXME: add comment
        self.moves:List['Move'] = []  # only the current hand
        self.randomGenerator:CountingRandom = CountingRandom(self)
        if wantedGame is None:
            wantedGame = str(int(self.randomGenerator.random() * 10 ** 9))
        self.first_point: Point
        self.last_point: Optional[Point]
        self.first_point, self.last_point = PointRange.from_string(wantedGame)
        # last_point is included like in 'N4'
        # here, last_point might be None meaning "open end" - it will be limited after loading ruleset
        self._point = Point(self.first_point)
        self.ruleset:Ruleset = ruleset
        self.__prevPoint = self._point
        self.__currentPoint = self._point
        self._prevPoint:Optional[Point] = None
        self.gameid:Optional[int] = gameid
        self.playOpen:bool = False
        self.autoPlay:bool = False
        self.divideAt:Optional[int] = None
        self.__lastDiscard:Optional[Tile] = None  # always uppercase
        # TODO: use Tile.none and remove assertions in message.py and otherwhere
        self.visibleTiles:Dict[Tile, int] = IntDict()
        self.discardedTiles:Dict[Tile, int] = IntDict(self.visibleTiles)
        # tile names are always lowercase
        self.dangerousTiles:List[Tuple[Set[Tile], str]] = []
        self.csvTags:List[str] = []
        self._setHandSeed()
        self.activePlayer:Optional[Player] = None
        self.__winner:Optional[Player] = None
        self._setGameId()
        self.__loadRuleset()
        # shift rules taken from the OEMC 2005 rules
        # 2nd round: S and W shift, E and N shift
        self.wall:Optional[Wall] = self.wallClass(self)  # type:ignore[arg-type]
        # FIXME:  wall nach PlayingGame verschieben?
        self.assignPlayers(names)  # also defines self.myself
        if self.belongsToGameServer():
            self.__shufflePlayers()
        for player in self.players:
            player.clearHand()

    @property
    def point(self) ->Point:
        """current position"""
        if self._point != self.__currentPoint:
            self.__prevPoint = self.__currentPoint
            self.__currentPoint = self._point
        return self._point

    @point.setter
    def point(self, value:Point) ->None:
        """go there"""
        if value.moveCount > 0:
            raise NotImplementedError('Game: changing position only accepts moveCount 0')
        # the while loop should only compare prevailing and rotated
        final_notRotated = value.notRotated
        value.notRotated = 0
        self._point.notRotated = 0
        while self.point < value: # here we use the getter for updating prevPoint
            self.rotateWinds()
        self._point.notRotated = final_notRotated

    @property
    def shouldSave(self) ->bool:
        """as a property"""
        return self.__shouldSave

    @shouldSave.setter
    def shouldSave(self, value:bool) ->None:
        """if activated, save start time"""
        if value and not self.__shouldSave:
            self.saveStartTime()
        self.__shouldSave = value

    @property
    def fullWallSize(self) ->int:
        """How many tiles we want to play with"""
        # the assertion for wallSize should not be done more often than needed: leave it in Wall()
        return int(Debug.wallSize) or elements.count(self.ruleset)

    @property
    def client(self) ->Optional['Client']:
        """hide weakref"""
        return self._client() if self._client else None

    @client.setter
    def client(self, value:Optional['Client']) ->None:
        """hide weakref"""
        if value:
            self._client = weakref.ref(value)
        else:
            self._client = None

    def clearHand(self) ->None:
        """empty all data"""
        while self.moves:
            _ = self.moves.pop()
            del _
        for player in self.players:
            player.clearHand()
        self.__winner = None
        self.prevActivePlayer = None
        self.dangerousTiles = []
        self.discardedTiles.clear()
        assert cast(IntDict, self.visibleTiles).count() == 0

    @property
    def lastDiscard(self) ->Optional[Tile]:
        """hide weakref"""
        return self.__lastDiscard

    @lastDiscard.setter
    def lastDiscard(self, value:Optional[Tile]) ->None:
        """hide weakref"""
        self.__lastDiscard = value
        if value is not None:
            assert isinstance(value, Tile), value
            if value.isExposed:
                raise ValueError(f'lastDiscard is exposed:{value}')

    @property
    def winner(self) ->Optional['Player']:
        """the name of the game server this game is attached to"""
        return self.__winner

    @winner.setter
    def winner(self, value:Optional['Player']) ->None:
        """the name of the game server this game is attached to"""
        if self.__winner != value:
            if self.__winner:
                self.__winner.invalidateHand()
            self.__winner = value
            if value:
                value.invalidateHand()

    def addCsvTag(self, tag:str, forAllPlayers:bool=False) ->None:
        """tag will be written to tag field in csv row"""
        if forAllPlayers or self.belongsToHumanPlayer():
            self.csvTags.append(f'{tag}/{self.point.prompt(self, withSeed=False)}')

    def isFirstHand(self) ->bool:
        """as the name says"""
        return self.point.is_in_first_hand()

    def _setGameId(self) ->None:
        """virtual"""
        raise NotImplementedError

    def close(self) ->None:
        """log off from the server"""
        self.wall = None
        self.lastDiscard = None
        if Options.gui:
            ParallelAnimationGroup.cancelAll()

    def playerByName(self, playerName:str) ->Optional[Player]:
        """return None or the matching player"""
        if playerName is None:
            return None
        for myPlayer in self.players:
            if myPlayer.name == playerName:
                return myPlayer
        logException(f'Move references unknown player {playerName}')
        return None

    def losers(self) ->List[Player]:
        """the 3 or 4 losers: All players without the winner"""
        return list(x for x in self.players if x is not self.__winner)

    def belongsToRobotPlayer(self) ->bool:
        """does this game instance belong to a robot player?"""
        return self.client is not None and self.client.isRobotClient()

    def belongsToHumanPlayer(self) ->bool:
        """does this game instance belong to a human player?"""
        return self.client is not None and self.client.isHumanClient()

    def belongsToGameServer(self) ->bool:
        """does this game instance belong to the game server?"""
        return self.client is not None and self.client.isServerClient()

    @staticmethod
    def isScoringGame() ->bool:
        """are we scoring a manual game?"""
        return False

    def belongsToPlayer(self) ->bool:
        """does this game instance belong to a player
        (as opposed to the game server)?"""
        return self.belongsToRobotPlayer() or self.belongsToHumanPlayer()

    def assignPlayers(self, playerNames:List[Tuple[Wind, str]]) ->None:
        """
        The server tells us the seating order and player names.

        @param playerNames: A list of 4 tuples. Each tuple holds wind and name.
        @type playerNames: The tuple contents must be C{str}
        @todo: Can we pass L{Players} instead of that tuple list?
        """
        if not self.players:
            self.players = Players()
            for idx in range(4):
                # append each separately: Until they have names, the current length of players
                # is used to assign one of the four walls to the player
                self.players.append(self.playerClass(
                    self, playerNames[idx][1]))
        for wind, name in playerNames:
            _ = self.players.byName(name)
            assert _
            _.wind = wind
        if self.client and self.client.name:
            _ = self.players.byName(self.client.name)
            assert _
            self.myself = _
        self.sortPlayers()

    def __shufflePlayers(self) ->None:
        """assign random seats to the players and assign winds"""
        self.players.sort(key=lambda x: x.name)
        self.randomGenerator.shuffle(self.players)
        for player, wind in zip(self.players, Wind.all4):
            player.wind = wind

    def __exchangeSeats(self) ->None:
        """execute seat exchanges according to the rules"""
        _ = {East: (), South: (South, West, East, North), West: (South, East), North: (West, East)}
        winds = _[self.point.prevailing]

        players = [self.players[x] for x in winds]
        pairs = [players[x:x + 2] for x in range(0, len(winds), 2)]
        for playerA, playerB in self._mustExchangeSeats(pairs):
            playerA.wind, playerB.wind = playerB.wind, playerA.wind

    def _mustExchangeSeats(self, pairs:List[List[Player]]) ->List[List[Player]]:
        """filter: which player pairs should really swap places?"""
        return pairs

    def sortPlayers(self) ->None:
        """sort by wind order. Place ourself at bottom (idx=0)"""
        self.players.sort(key=lambda x: x.wind)
        self.activePlayer = self.players[East]
        if Internal.scene:
            if self.belongsToHumanPlayer():
                while self.players[0] != self.myself:
                    self.players = Players(self.players[1:] + self.players[:1])
                for idx, player in enumerate(self.players):
                    assert self.wall
                    player.front = cast('UIWall', self.wall)[idx]
                    if hasattr(player, 'sideText'):
                        if player.sideText.board is not player.front:
                            player.sideText.animateNextChange = True
                        player.sideText.board = player.front

    def saveStartTime(self) ->None:
        """save starttime for this game"""
        starttime = datetime.datetime.now().replace(microsecond=0).isoformat()
        args:List[Union[str, int, float]] = list([starttime, self.seed, int(self.autoPlay),
                     self.ruleset.rulesetId])
        args.extend([p.nameid for p in self.players])
        assert self.gameid
        args.append(self.gameid)
        Query("update game set starttime=?,seed=?,autoplay=?,"
              "ruleset=?,p0=?,p1=?,p2=?,p3=? where id=?", tuple(args))

    def __loadRuleset(self) ->None:
        """use a copy of ruleset for this game, reusing an existing copy"""
        self.ruleset.load()
        if Internal.db:
            # only if we have a DB open. False in scoringtest.py
            query = Query(
                'select id from ruleset where id>0 and hash=?',
                (self.ruleset.hash,))
            if query.records:
                # reuse that ruleset
                self.ruleset.rulesetId = query.records[0][0]
            else:
                # generate a new ruleset
                self.ruleset.save()
        rule_last_point = Point(f'{self.seed}/{Wind.all4[self.ruleset.minRounds - 1]}4')
        if self.last_point is None:
            self.last_point = rule_last_point
        else:
            self.last_point = min(rule_last_point, self.last_point)

    @property
    def seed(self) ->int:
        """extract it from wantedGame. Set wantedGame if empty."""
        return self.first_point.seed

    def _setHandSeed(self) ->None:
        """set seed to a reproducible value, independent of what happened
        in previous hands/rounds.
        This makes it easier to reproduce game situations
        in later hands without having to exactly replay all previous hands"""
        self.randomGenerator.seed(self.seed * self.point.seedFactor())

    def prepareHand(self) ->None:
        """prepare a game hand"""
        self.clearHand()
        if self.finished():
            self.close()

    def initHand(self) ->None:
        """directly before starting"""
        self.dangerousTiles = []
        self.discardedTiles.clear()
        assert cast(IntDict, self.visibleTiles).count() == 0
        if Internal.scene:
            # TODO: why not self.scene?
            Internal.scene.prepareHand()
        self._setHandSeed()

    def saveHand(self) ->None:
        """save hand to database,
        update score table and balance in status line"""
        self.__payHand()
        self._saveScores()
        self.point.endHand()

    def _saveScores(self) ->None:
        """save computed values to database,
        update score table and balance in status line"""
        scoretime = datetime.datetime.now().replace(microsecond=0).isoformat()
        logMessage = ''
        for player in self.players:
            if player.hand:
                manualrules = '||'.join(x.rule.name
                                        for x in player.hand.usedRules)
            else:
                manualrules = i18n('Score computed manually')
            assert self.gameid
            Query(
                "INSERT INTO SCORE "  # pylint:disable=consider-using-f-string
                "(game,hand,data,manualrules,player,scoretime,won,prevailing,"
                "wind,points,payments, balance,rotated,notrotated) "
                "VALUES(%d,%d,?,?,%d,'%s',%d,'%s','%s',%d,%d,%d,%d,%d)" %
                (self.gameid, self.point.handCount, player.nameid,
                 scoretime, int(player == self.__winner),
                 self.point.prevailing, player.wind,
                 player.handTotal, player.payment, player.balance,
                 self.point.rotated, self.point.notRotated),
                (player.hand.string, manualrules))
            logMessage += (f"{str(player)[:12]:<12} {player.handTotal:>4} {player.balance:>5} "
                          f"{'WON' if player == self.winner else '   '} | ")
            for usedRule in player.hand.usedRules:
                rule = usedRule.rule
                if rule.score.limits:
                    self.addCsvTag(rule.name.replace(' ', ''))
        if Debug.scores:
            self.debug(logMessage)

    def maybeRotateWinds(self) ->bool:
        """rules which make winds rotate"""
        result = [x for x in self.ruleset.filterRules('rotate') if x.rotate(self)]  # type:ignore[attr-defined]
        if result:
            if Debug.explain:
                if not self.belongsToRobotPlayer():
                    self.debug(','.join(x.name for x in result), showPrevPoint=True)
            self.rotateWinds()
            if self.finished():
                self.finish_in_db()
        return bool(result)

    def rotateWinds(self) ->None:
        """rotate winds, exchange seats. If finished, update database"""
        self.point.rotate()
        if not self.finished() and not self.belongsToPlayer():
            # the game server already told us the new placement and winds
            winds = [player.wind for player in self.players]
            winds = winds[3:] + winds[0:3]
            for idx, newWind in enumerate(winds):
                self.players[idx].wind = newWind
            if self.point.rotated == 0:
                # exchange seats between rounds
                self.__exchangeSeats()
            if Internal.scene:
                with AnimationSpeed(Speeds.windDisc):
                    cast('UIWall', self.wall).showWindDiscs()

    def finish_in_db(self) ->None:
        """write endtime into db"""
        endtime = datetime.datetime.now().replace(
            microsecond=0).isoformat()
        assert self.gameid
        with Internal.db as transaction:
            transaction.execute(
                f'UPDATE game set endtime = "{endtime}" where id = {int(self.gameid)}')

    def debug(self, msg:str, btIndent:Optional[int]=None,
        showPrevPoint:bool=False, showStack:bool=False) ->None:
        """
        Log a debug message.

        @param msg: The message.
        @type msg: A string.
        @param btIndent: If given, message is indented by
        depth(backtrace)-btIndent
        @type btIndent: C{int}
        @param showPrevPoint: If True, do not use current point but previous
        @type showPrevPoint: C{bool}
        """
        if self.belongsToRobotPlayer():
            prefix = 'R'
        elif self.belongsToHumanPlayer():
            prefix = 'C'
        elif self.belongsToGameServer():
            prefix = 'S'
        else:
            logDebug(msg, btIndent=btIndent)
            return
        point = self.__prevPoint if showPrevPoint else self.point
        point = self.point
        assert point
        point_str = point.prompt(self, withMoveCount=True)
        logDebug(
            f'{prefix}{point_str}: {msg}',
            withGamePrefix=False,
            btIndent=btIndent,
            showStack=showStack)

    @staticmethod
    def __getName(playerid:int) ->str:
        """get name for playerid
        """
        try:
            return Players.allNames[playerid]
        except KeyError:
            return i18n('Player %1 not known', playerid)

    @classmethod
    def _loadGameRecord(cls, gameid:int) ->Optional[Any]:
        """load and sanitize"""
        record = Query(
            "select {fields} from game where id = ?",
            (gameid,), fields='p0,p1,p2,p3,ruleset,id,seed').tuple()
        if record:
            return record._replace(ruleset=int(record.ruleset) or 1)
        return None

    @classmethod
    def _loadLastHand(cls, gameid:int) ->int:
        """load or invent"""
        records = Query(
            "select {fields} from score "
            "where game=? "
                "and hand=(select max(hand) from score where game=?) ",
            (gameid, gameid), fields='hand').records
        if records:
            return int(records[0][0])
        return 0

    @classmethod
    def _loadScores(cls, qGame: Any, hand:int) ->Any:
        """If the server saved a score entry but our client
           did not, we get no record here. Should we try to fix this or
           exclude such a game from the list of resumable games?"""
        scoreFields = 'player, wind, balance, won, prevailing'
        qScores = Query(
            "select {fields} from score "
            "where game=? and hand=?",
            (qGame.id, hand), fields=scoreFields).tuples()  # type: ignore
        if not qScores:
            # this should normally not happen
            my_class = namedtuple('_', scoreFields)  # type: ignore
            qScores = list(
                my_class(qGame[wind], wind.char, 0, False, East.char)  # type:ignore
                for wind in Wind.all4)  # type: ignore[call-arg]
        if len(qScores) != 4:
            logError(
                f'game {int(qGame.id)}: last hand should have 4 score records, found {len(qScores)}')
        if len({x.prevailing for x in qScores}) != 1:
            logError(f'game {qGame.id} inconsistent: '
                     f'All score records for the same hand must have the same prevailing wind')
        return qScores

    @classmethod
    def loadFromDB(cls, gameid:int, client:Optional['Client']=None) ->Optional[Union['Game', 'ServerGame']]:
        """load game by game id and return a new Game instance"""
        # TODO would be nice to use cls in result annotation, but how?
        Internal.logPrefix = 'S' if Internal.isServer else 'C'
        qGame = cls._loadGameRecord(gameid)
        if qGame is None:
            return None
        ruleset = Ruleset.cached(qGame.ruleset)
        Players.load()  # we want to make sure we have the current definitions
        hand = cls._loadLastHand(qGame.id)
        qScores = cls._loadScores(qGame, hand)

        players = list((x.wind, Game.__getName(x.player)) for x in qScores)

        last_point = Point(gameid)
        last_point.endHand()

        resume_at = f'{qGame.seed}/{last_point.full_str()}'

        # create the game instance
        game = cls(players, ruleset, gameid=gameid, client=client, wantedGame=resume_at)

        for qScore in qScores:
            player = game.players.byId(qScore.player)
            if not player:
                logError(
                    f'game {int(gameid)} inconsistent: player {qScore.player} missing in game table')
            else:
                player.getsPayment(qScore.balance)
            if qScore.won:
                game.winner = player
        game.maybeRotateWinds()
        game.sortPlayers()
        with AnimationSpeed(Speeds.windDisc):
            assert game.wall
            animateAndDo(game.wall.decorate4)
        return game

    def finished(self) ->bool:
        """The game is over after minRounds completed rounds. Also,
        check if we reached the second point defined by --game.
        If we did, the game is over too"""
        return self.point > self.last_point

    def __payHand(self) ->None:
        """pay the scores"""
        # pylint: disable=too-many-branches
        # too many branches
        winner = self.__winner
        if winner:
            winner.wonCount += 1
            guilty = winner.usedDangerousFrom
            payAction = self.ruleset.findUniqueOption('payforall') if guilty else None
            if payAction:
                assert isinstance(guilty, Player)  # mypy should be able to infer this
                if Debug.dangerousGame:
                    self.debug(f'{self.point}: winner {winner}. {guilty} pays for all')
                guilty.hand.usedRules.append(UsedRule(payAction))
                score = winner.handTotal
                score = score * 6 if winner.wind == East else score * 4
                guilty.getsPayment(-score)
                winner.getsPayment(score)
                return

        for player1 in self.players:
            if Debug.explain:
                if not self.belongsToRobotPlayer():
                    self.debug(f'{player1}: {player1.hand.string}')
                    for line in player1.hand.explain():
                        self.debug(f'   {line}')
            for player2 in self.players:
                if id(player1) != id(player2):
                    if East in (player1.wind, player2.wind):
                        efactor = 2
                    else:
                        efactor = 1
                    if player2 != winner:
                        player1.getsPayment(player1.handTotal * efactor)
                    if player1 != winner:
                        player1.getsPayment(-player2.handTotal * efactor)

    def lastMoves(self, only:Optional[Iterable['Message']]=None,
        without:Optional[Iterable['Message']]=None,
        withoutNotifications:bool=False) ->Generator['Move', None, None]:
        """filters and yields the moves in reversed order"""
        for idx in range(len(self.moves) - 1, -1, -1):
            move = self.moves[idx]
            if withoutNotifications and move.notifying:
                continue
            if only:
                if move.message in only:
                    yield move
            elif without:
                if move.message not in without:
                    yield move
            else:
                yield move

    def throwDices(self) ->None:
        """set random living and kongBox
        sets divideAt: an index for the wall break"""
        breakWall = self.randomGenerator.randrange(4)
        assert self.wall
        sideLength = len(self.wall) // 4
        # use the sum of four dices to find the divide
        self.divideAt = breakWall * sideLength + \
            sum(self.randomGenerator.randrange(1, 7) for idx in range(4))
        if self.divideAt % 2 == 1:
            self.divideAt -= 1
        self.divideAt %= len(self.wall)


class PlayingGame(Game):
    """this game is played using the computer"""

    # pylint: disable=too-many-instance-attributes,useless-suppression
    # pylint 2.16.2 is erratic here, sometimes it warns, sometimes not

    playerClass = PlayingPlayer

    def __init__(self, names:List[Tuple['Wind', str]], ruleset:Ruleset, gameid:Optional[int]=None,
                 wantedGame:Optional[str]=None, client:Optional['Client']=None,
                 playOpen:bool=False, autoPlay:bool=False) ->None:
        """a new game instance, comes from database if gameid is set"""
        self.__activePlayer:Optional[PlayingPlayer] = None
        self.prevActivePlayer:Optional[PlayingPlayer] = None
        self.defaultNameBrush = None
        super().__init__(names, ruleset,
                      gameid, wantedGame=wantedGame, client=client)
        self.players[East].lastSource = TileSource.East14th
        self.playOpen = playOpen
        self.autoPlay = autoPlay
        if self.belongsToHumanPlayer():
            myself = cast(PlayingPlayer, self.myself)
            myself.voice = Voice.locate(myself.name)
            if myself.voice:
                if Debug.sound:
                    logDebug(f'myself {myself.name} gets voice {myself.voice}')
            else:
                if Debug.sound:
                    logDebug(f'myself {myself.name} gets no voice')

    def writeCsv(self) ->None:
        """write game summary to Options.csv"""
        if self.finished() and Options.csv:
            gameWinner = max(self.players, key=lambda x: x.balance)
            if Debug.process and sys.platform != 'win32':
                self.csvTags.append(f'MEM:{resource.getrusage(resource.RUSAGE_SELF).ru_maxrss}')  # pylint:disable=possibly-used-before-assignment
            _ = CsvRow.Fields
            row = [''] * CsvRow.Fields.PLAYERS
            row[_.GAME] = str(self.seed)
            row[_.RULESET] = self.ruleset.name
            row[_.AI] = Options.AI
            row[_.COMMIT] = gitHead()
            row[_.PY_VERSION] = '{}.{}'.format(*sys.version_info[:2])  # pylint:disable=consider-using-f-string
            row[_.TAGS] = ','.join(self.csvTags)
            for player in sorted(self.players, key=lambda x: x.name):
                row.append(player.name)
                row.append(player.balance)
                row.append(player.wonCount)
                row.append(1 if player == gameWinner else 0)
            CsvRow(row).write()

    def close(self) ->Any:
        """log off from the server and return a Deferred"""
        Game.close(self)
        self.writeCsv()
        Internal.autoPlay = False  # do that only for the first game
        if self.client:
            client = self.client
            self.client = None
            result = client.logout()
        else:
            result = succeed(None)
        return result

    def _setGameId(self) ->None:
        """do nothing, we already went through the game id reservation"""

    @property  # type:ignore[override]
    def activePlayer(self) ->PlayingPlayer:
        """the turn is on this player"""
        result = self.__activePlayer
        assert result
        return result

    @activePlayer.setter
    def activePlayer(self, player:PlayingPlayer) ->None:
        """the turn is on this player"""
        if self.__activePlayer != player:
            self.prevActivePlayer = self.__activePlayer
            if self.prevActivePlayer:
                self.prevActivePlayer.hidePopup()
            self.__activePlayer = player
            if Internal.scene:  # mark the name of the active player in blue
                for _ in self.players:
                    _.colorizeName()

    def prepareHand(self) ->None:
        """prepares the next hand"""
        Game.prepareHand(self)
        if not self.finished():
            self.sortPlayers()
            self.hidePopups()
            self._setHandSeed()
            assert self.wall
            self.wall.build(shuffleFirst=True)

    def hidePopups(self) ->None:
        """hide all popup messages"""
        for player in self.players:
            player.hidePopup()

    def saveStartTime(self) ->None:
        """write a new entry in the game table with the selected players"""
        if not self.gameid:
            # in server.__prepareNewGame, gameid is None here
            return
        records = Query("select seed from game where id=?", (
            self.gameid,)).records
        assert records, f'self.gameid: {self.gameid}'
        seed = records[0][0]

        if not Internal.isServer and self.client:
            assert self.client.connection
            host = self.client.connection.url
        else:
            host = None

        if seed in ('proposed', host):
            # we reserved the game id by writing a record with seed == host
            super().saveStartTime()

    def _saveScores(self) ->None:
        """save computed values to database, update score table
        and balance in status line"""
        if self.shouldSave:
            if self.belongsToRobotPlayer():
                assert False, 'shouldSave must not be True for robot player'
            super()._saveScores()

    def nextPlayer(self, current:Optional[PlayingPlayer]=None) ->PlayingPlayer:
        """return the player after current or after activePlayer"""
        if not current:
            current = self.activePlayer
        pIdx = self.players.index(current)
        return cast(PlayingPlayer, self.players[(pIdx + 1) % 4])

    def nextTurn(self) ->None:
        """move activePlayer"""
        self.activePlayer = self.nextPlayer()

    def _concealedTileName(self, tileName:Tile) ->Tile:
        """tileName has been discarded, by which name did we know it?"""
        player = self.activePlayer
        if player != self.myself and not self.playOpen:
            # we are human and server tells us another player discarded
            # a tile. In our game instance, tiles in handBoards of other
            # players are unknown
            player.makeTileKnown(tileName)
            result = Tile.none
        else:
            result = tileName
        if tileName not in player.concealedTiles:
            raise ValueError(f"I am {self.myself.name if self.myself else 'None'}. Player {player.name} "
                             f"is told to show discard of tile {result} "
                             f"but does not have it, he has {player.concealedTiles}")
        return result

    def hasDiscarded(self, player:PlayingPlayer, tile:Tile) ->None:
        """discards a tile from a player board"""
        assert isinstance(tile, Tile)
        if player != self.activePlayer:
            raise ValueError(f'Player {player} discards but {self.activePlayer} is active')
        self.discardedTiles[tile.exposed] += 1
        player.discarded.append(tile)
        player.lastTile = Tile.none
        self._concealedTileName(tile)
        # the above has side effect, needs to be called
        if Internal.scene:
            assert player.handBoard
            cast('PlayingHandBoard', player.handBoard).discard(tile)
        self.lastDiscard = tile
        player.removeConcealedTile(self.lastDiscard)
        if any(tile.exposed in x[0] for x in self.dangerousTiles):
            self.computeDangerous()
        else:
            self._endWallDangerous()

    def saveHand(self) ->None:
        """server told us to save this hand"""
        for player in self.players:
            handWonMatches = player.hand.won == (player == self.winner)
            assert handWonMatches, f'hand.won:{player.hand.won} winner:{player == self.winner}'
        super().saveHand()

    def _mustExchangeSeats(self, pairs:List[List[Player]]) ->List[List[Player]]:
        """filter: which player pairs should really swap places?"""
        # if we are a client in a remote game, the server swaps and tells
        # us the new places
        return [] if self.belongsToPlayer() else pairs

    def assignVoices(self) ->None:
        """now we have all remote user voices"""
        assert self.belongsToHumanPlayer()
        available = Voice.availableVoices()[:]
        # available is without transferred human voices
        for player in self.players:
            if player.voice and player.voice.oggFiles():
                # remote human player sent her voice, or we are human
                # and have a voice
                if Debug.sound and player != self.myself:
                    logDebug(f'{player.name} got voice from opponent: {player.voice}')
            else:
                player.voice = Voice.locate(player.name)
                if player.voice:
                    if Debug.sound:
                        logDebug(f'{player.name} has own local voice {player.voice}')
            if player.voice:
                for voice in Voice.availableVoices():
                    if (voice in available
                            and voice.md5sum == player.voice.md5sum):
                        # if the local voice is also predefined,
                        # make sure we do not use both
                        available.remove(voice)
        # for the other players use predefined voices in preferred language.
        # Only if we do not have enough predefined voices, look again in
        # locally defined voices
        predefined = [x for x in available if x.language() != 'local']
        predefined.extend(available)
        for player in self.players:
            if player.voice is None and predefined:
                player.voice = predefined.pop(0)
                if Debug.sound:
                    logDebug(
                        f'{player.name} gets one of the still available voices {player.voice}')

    def dangerousFor(self, forPlayer:PlayingPlayer, tile:Tile) ->List[str]:
        """return a list of explaining texts if discarding tile
        would be Dangerous game for forPlayer. One text for each
        reason - there might be more than one"""
        assert isinstance(tile, Tile), tile
        tile = tile.exposed
        result:List[str] = []
        for dang, txt in self.dangerousTiles:
            if tile in dang:
                result.append(txt)
        for player in forPlayer.others():
            for dang, txt in player.dangerousTiles:
                if tile in dang:
                    result.append(txt)
        return result

    def computeDangerous(self, playerChanged:Optional[PlayingPlayer]=None) ->None:
        """recompute gamewide dangerous tiles. Either for playerChanged or
        for all players"""
        self.dangerousTiles = []
        if playerChanged:
            playerChanged.findDangerousTiles()
        else:
            for player in self.players:
                player.findDangerousTiles()
        self._endWallDangerous()

    def _endWallDangerous(self) ->None:
        """if end of living wall is reached, declare all invisible tiles
        as dangerous"""
        assert self.wall
        if len(self.wall.living) <= 5:
            allTiles = [x for x in elements.occurrence if not x.isBonus]
            for tile in allTiles:
                assert isinstance(tile, Tile), tile
            # see https://www.logilab.org/ticket/23986
            invisibleTiles = {x for x in allTiles if x not in self.visibleTiles}
            msg = i18n('Short living wall: Tile is invisible, hence dangerous')
            self.dangerousTiles = [x for x in self.dangerousTiles if x[1] != msg]
            self.dangerousTiles.append((invisibleTiles, msg))



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime
from typing import Tuple, Optional, TYPE_CHECKING, List, cast, Any

from qt import QPointF, QRectF, QDialogButtonBox
from qt import QGraphicsRectItem, QGraphicsSimpleTextItem
from qt import QPushButton, QMessageBox, QComboBox


from common import Internal, isAlive, Debug
from tile import MeldList
from wind import Wind
from tilesource import TileSource
from animation import animate
from log import logException, logDebug, logWarning, i18n
from query import Query
from uitile import UITile, UIMeld
from board import WindLabel
from game import Game
from games import Games
from hand import Hand
from handboard import HandBoard, TileAttr
from player import Player, Players
from visible import VisiblePlayer
from tables import SelectRuleset
from uiwall import UIWall, SideText
from guiutil import decorateWindow, BlockSignals, rotateCenter, sceneRotation
from mi18n import i18nc
from util import callers

if TYPE_CHECKING:
    from qt import QWidget, QEvent, QGraphicsSceneDragDropEvent, QGraphicsItem, QGraphicsItemGroup, QObject
    from tile import Tile, Meld
    from board import MimeData, SelectorBoard
    from scoringdialog import RuleBox
    from scene import ScoringScene
    from rule import Score, UsedRule, Ruleset, Rule
    from client import Client

def scoringScene() ->'ScoringScene':
    """shortcut"""
    result = cast('ScoringScene', Internal.scene)
    assert result
    return result

class SwapDialog(QMessageBox):

    """ask the user if two players should change seats"""

    def __init__(self, swappers:List[Player], parent:Optional['QWidget']=None) ->None:
        super().__init__(parent)
        decorateWindow(self, i18nc("@title:window", "Swap Seats"))
        self.setText(
            i18n("By the rules, %1 and %2 should now exchange their seats. ",
                 swappers[0].name, swappers[1].name))
        self.yesAnswer = QPushButton(i18n("&Exchange"))
        self.addButton(self.yesAnswer, QMessageBox.ButtonRole.YesRole)
        self.noAnswer = QPushButton(i18n("&Keep seat"))
        self.addButton(self.noAnswer, QMessageBox.ButtonRole.NoRole)


class ScoringComboBox(QComboBox):

    """with flag manualSelect"""

    def __init__(self, parent:Optional['QWidget']=None):
        super().__init__(parent)
        self.manualSelect = False

class SelectPlayers(SelectRuleset):

    """a dialog for selecting four players. Used only for scoring game."""

    def __init__(self) ->None:
        super().__init__()
        Players.load()
        decorateWindow(self, i18nc("@title:window", "Select four players"))
        self.names:List[str]
        self.nameWidgets:List[ScoringComboBox] = []
        for idx, wind in enumerate(Wind.all4):
            cbName = ScoringComboBox()
            # increase width, we want to see the full window title
            cbName.setMinimumWidth(350)  # is this good for all platforms?
            cbName.addItems(list(Players.humanNames.values()))
            self.grid.addWidget(cbName, idx + 1, 1)
            self.nameWidgets.append(cbName)
            self.grid.addWidget(WindLabel(wind), idx + 1, 0)
            cbName.currentIndexChanged.connect(self.slotValidate)

        # scoring games have seed 0
        # a scoring game which only just started or threw an exception might have p0,p1,p2,p3 = null
        query = Query(
            "select id,p0,p1,p2,p3 from game where seed=0 and p0 is not null order by id desc limit 1")
        if query.records:
            record = query.record()
            max_game_id = record[0]
            with BlockSignals(self.nameWidgets):
                for cbName, playerId in zip(self.nameWidgets, record[1:]):
                    if playerId not in Players.humanNames:
                        logException(f'database {Internal.db.path} is inconsistent: '
                                     f'game {max_game_id}: player with id {playerId} '
                                     f'is in game but not in player')
                    playerName = Players.humanNames[playerId]
                    playerIdx = cbName.findText(playerName)
                    if playerIdx >= 0:
                        cbName.setCurrentIndex(playerIdx)
        self.slotValidate()

    def showEvent(self, unusedEvent:Optional['QEvent']) ->None:
        """start with player 0"""
        self.nameWidgets[0].setFocus()

    def __selectedNames(self) ->set[str]:
        """A set with the currently selected names"""
        return {cbName.currentText() for cbName in self.nameWidgets}

    def slotValidate(self) ->None:
        """try to find 4 different players and update status of the Ok button"""
        changedCombo = self.sender()
        if not isinstance(changedCombo, ScoringComboBox):
            changedCombo = self.nameWidgets[0]
        changedCombo.manualSelect = True
        allNames = set(Players.humanNames.values())
        unusedNames = allNames - self.__selectedNames()
        with BlockSignals(self.nameWidgets):
            used = {x.currentText() for x in self.nameWidgets if x.manualSelect}
            for combo in self.nameWidgets:
                if not combo.manualSelect:
                    if combo.currentText() in used:
                        comboName = unusedNames.pop()
                        combo.clear()
                        combo.addItems([comboName])
                        used.add(combo.currentText())
            for combo in self.nameWidgets:
                comboName = combo.currentText()
                combo.clear()
                combo.addItems([comboName])
                combo.addItems(sorted(
                    allNames - self.__selectedNames() - {comboName}))
                combo.setCurrentIndex(0)
        ok_button = self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        if ok_button:
            ok_button.setEnabled(len(self.__selectedNames()) == 4)
        self.names = [cbName.currentText() for cbName in self.nameWidgets]


class ScoringTileAttr(TileAttr):

    """Tile appearance is different in a ScoringHandBoard"""

    def _setDark(self) ->bool:
        """should the tile appear darker?"""
        return bool(self.yoffset) or self.tile.isConcealed

    def _setFocusable(self, hand:'HandBoard', meld:'Meld', idx:Optional[int]) ->bool:
        """in a scoring handboard, only the first tile of a meld is focusable"""
        return idx == 0


class ScoringHandBoard(HandBoard):

    """a board showing the tiles a player holds"""
    tileAttrClass = ScoringTileAttr

    def __init__(self, player:'ScoringPlayer') ->None:
        self.__moveHelper:Optional['QGraphicsItemGroup'] = None
        self.uiMelds:List[UIMeld] = []  # uiMelds and uiTiles are administered independently
        super().__init__(player)
        self.player:'ScoringPlayer'

    def meldVariants(self, tile:UITile, forLowerHalf:bool) ->MeldList:
        """Kong might have variants"""
        result = MeldList()
        uimeld = self.uiMeldWithTile(tile)
        if not uimeld:
            logWarning(f'ScoringHandBoard.meldVariants: no meld found for {tile}')
            return result
        meld = uimeld.meld
        result.append(meld.concealed if forLowerHalf else meld.exposed)
        if len(meld) == 4:
            if forLowerHalf:
                result.append(meld.declared)
            else:
                result.append(meld.exposedClaimed)
        return result

    def mapMouseTile(self, uiTile:UITile) ->UITile:
        """map the pressed tile to the wanted tile. For melds, this would
        be the first tile no matter which one is pressed"""
        uiMeld = self.uiMeldWithTile(uiTile)
        assert uiMeld
        return uiMeld[0]

    def uiMeldWithTile(self, uiTile:UITile, remove:bool=False) ->UIMeld:
        """return the meld with uiTile, optionally removing it"""
        for idx, myMeld in enumerate(self.uiMelds):
            if uiTile in myMeld:
                uiMeld = self.uiMelds[idx]
                if remove:
                    if Debug.uitiles:
                        logDebug(f'{callers()}.{self.debug_name()}.uiMeldWithTile({uiTile}) removes {uiMeld}')
                    del self.uiMelds[idx]
                    self.player.removeMeld(uiMeld.meld)
                    if Debug.uitiles:
                        logDebug(f'{self.debug_name()}.loseMeld({uiTile}) removed {uiMeld.meld} from {self.player}')
                    self.sync()
                return myMeld
        return UIMeld(uiTile)

    def findUIMeld(self, meld:Optional['Meld']) ->UIMeld:
        """find the first UIMeld matching the logical meld"""
        for result in self.uiMelds:
            if result.meld == meld:
                return result
        logWarning(f'Scoring game: cannot find UIMeld for {meld}')
        if self.uiMelds:
            # when could this happen?
            logWarning(f'XXXXXXXXXXXXXX findUIMeld returning uiMelds[1]: {self.uiMelds[1]}')
            return self.uiMelds[1] # FIXME why not 0 ?
        raise ValueError('Scoring Game: findUIMeld() in absence of any uiMelds')

    def loseMeld(self, uiTile:UITile, meld:'Meld') ->UIMeld:  # pylint: disable=unused-argument
        """loses a UIMeld. Only first uiTile is given"""
        assert isinstance(uiTile, UITile), uiTile
        uiMeld = self.uiMeldWithTile(uiTile, remove=True)
        scoringScene().handSelectorChanged(self)
        return uiMeld

    def sync(self, adding:Optional[List[UITile]]=None) ->None:
        """place all tiles in ScoringHandBoard"""
        col = [0.0, 0.0]
        boni:List[UITile] = []
        for uiMeld in self.uiMelds:
            if uiMeld[0].tile.isBonus:
                boni.append(uiMeld[0])
                continue
            row = 0 if uiMeld.meld[0].isExposed else 1
            for idx, tile in enumerate(uiMeld):
                tile.level = 0
                tile.setBoard(self, col[row], row)
                # FIXME Folgendes UIMeld sollte sich selber so einstellen
                tile.focusable = idx == 0
                tile.dark = bool(row) or tile.tile.isConcealed
                col[row] += 1
            col[row] += self.exposedMeldDistance
        self._placeBonusTiles(col, boni)
        # self.checkTiles() does not work here

    def dragMoveEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """allow dropping of uiTile from ourself only to other state (open/concealed)"""
        if not event:
            return
        uiTile = cast('MimeData', event.mimeData()).uiTile
        localY = self.mapFromScene(QPointF(event.scenePos())).y()
        centerY = self.rect().height() / 2.0
        newLowerHalf = localY >= centerY
        noMansLand = centerY / 6
        if -noMansLand < localY - centerY < noMansLand and not uiTile.isBonus:
            doAccept = False
        elif uiTile.board != self:
            doAccept = True
        elif uiTile.isBonus:
            doAccept = False
        else:
            oldLowerHalf = False
            if uiTile.board.isHandBoard:
                oldLowerHalf = uiTile in cast('HandBoard', uiTile.board).lowerHalfTiles()
            doAccept = oldLowerHalf != newLowerHalf
        event.setAccepted(doAccept)

    def dropEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """drop into this handboard"""
        if not event:
            return
        uiTile = cast('MimeData', event.mimeData()).uiTile
        forLowerHalf = self.mapFromScene(
            QPointF(event.scenePos())).y() >= self.rect().height() / 2.0
        if self.dropMeld(uiTile, forLowerHalf):
            event.accept()
        else:
            event.ignore()
        self._noPen()

    def dropMeld(self, uiTile:UITile, forLowerHalf:bool) ->bool:
        """drop UIMeld containing uiTile into lower or upper half of our hand"""
        senderBoard = cast('SelectorBoard', uiTile.board)
        assert senderBoard
        newMeld = senderBoard.chooseVariant(uiTile, forLowerHalf)
        if not newMeld:
            return False
        uiMeld = senderBoard.loseMeld(uiTile, newMeld)
        for uitile, tile in zip(uiMeld, newMeld):
            uitile.change_name(tile)
        self.uiMelds.append(uiMeld)
        if Debug.uitiles:
            logDebug(f'{self.debug_name()}.dropMeld({uiTile}) appends{uiMeld}')
        assert self.player
        self.player.addMeld(uiMeld.meld)
        if Debug.uitiles:
            logDebug(f'{self.debug_name()}.dropMeld({uiTile}) added {uiMeld.meld} to {self.player}')
        self.sync()
        self.hasLogicalFocus = senderBoard == self or senderBoard.empty
        self.checkTiles()
        senderBoard.autoSelectTile()
        senderBoard.checkTiles()
        if senderBoard is not self and senderBoard.isHandBoard:
            scoringScene().handSelectorChanged(cast('HandBoard', senderBoard))
        scoringScene().handSelectorChanged(self)
        animate()
        self.checkTiles()
        return True

    def focusRectWidth(self) ->int:
        """how many tiles are in focus rect? We want to focus
        the entire meld"""
        focus = self.focusTile
        if not focus:
            return 1
        meld = self.uiMeldWithTile(focus)
        if not meld:
            logWarning(f'ScoringBoard.focusRectWidth: there is no meld with focus tile {focus}')
            return 1
        if Debug.focusable:
            logDebug(f'{callers()}.{self}.focusRectWidth for focusTile {self.focusTile} '
                     f'found UIMeld {meld} with tiles {list(x for x in meld)}')
        return len(meld)

    def addUITile(self, uiTile:UITile) ->None:
        super().addUITile(uiTile)
        self.showMoveHelper()

    def removeUITile(self, uiTile:UITile) ->None:
        super().removeUITile(uiTile)
        self.showMoveHelper()

    def showMoveHelper(self, visible:Optional[bool]=None) ->None:
        """show help text In empty HandBoards"""
        if visible is None:
            visible = self.empty
        if self.__moveHelper and not isAlive(self.__moveHelper):
            return
        if visible:
            if not self.__moveHelper:
                splitter = QGraphicsRectItem(self)
                hbCenter = self.rect().center()
                splitter.setRect(
                    hbCenter.x() * 0.5,
                    hbCenter.y(),
                    hbCenter.x() * 1,
                    1)
                helpItems:List['QGraphicsItem'] = [splitter]
                for name, yFactor in [(i18n('Move Exposed Tiles Here'), 0.5),
                                      (i18n('Move Concealed Tiles Here'), 1.5)]:
                    helper = QGraphicsSimpleTextItem(name, self)
                    helper.setScale(3)
                    nameRect = QRectF()
                    nameRect.setSize(
                        helper.mapToParent(helper.boundingRect()).boundingRect().size())
                    center = QPointF(hbCenter)
                    center.setY(center.y() * yFactor)
                    _ = center
                    _ -= nameRect.center()
                    helper.setPos(_)
                    if sceneRotation(self) == 180:
                        rotateCenter(helper, 180)
                    helpItems.append(helper)
                scene = self.scene()
                if scene:
                    self.__moveHelper = scene.createItemGroup(helpItems)
            if self.__moveHelper:
                self.__moveHelper.setVisible(True)
        else:
            if self.__moveHelper:
                self.__moveHelper.setVisible(False)


class ScoringPlayer(VisiblePlayer, Player):

    """Player in a scoring game"""

    def __init__(self, game:'ScoringGame', name:str) ->None:
        self.handBoard = None  # because Player.init calls clearHand()
        self.manualRuleBoxes:List['RuleBox'] = []
        Player.__init__(self, game, name)
        VisiblePlayer.__init__(self)
        self.handBoard = ScoringHandBoard(self)
        self.game:'ScoringGame'

    def clearHand(self) ->None:
        """clears attributes related to current hand"""
        super().clearHand()
        if self.game and self.game.wall:
            assert self.game
            assert self.game.wall
            # is None while __del__
            self.front = self.game.wall[self.idx]
            if hasattr(self, 'sideText'):
                self.sideText.board = self.front
        if isAlive(self.handBoard):
            assert (_ := cast(ScoringHandBoard, self.handBoard))
            _.setEnabled(True)
            _.showMoveHelper()
            _.uiMelds = []
        self.manualRuleBoxes = []

    def explainHand(self) ->'Hand':
        """return the hand to be explained"""
        return self.hand

    @property
    def handTotal(self) ->int:
        """the hand total of this player"""
        if self.hasManualScore():
            assert (_ := scoringScene().scoringDialog)
            spValue = _.spValues[self.idx]
            return spValue.value()
        return self.hand.total()

    def hasManualScore(self) ->bool:
        """True if no tiles are assigned to this player"""
        if _ := scoringScene().scoringDialog:
            return _.spValues[self.idx].isEnabled()
        return False

    def refreshManualRules(self, sender:Optional['QObject']=None) ->None:
        """update status of manual rules"""
        if not self.handBoard:
            # might happen at program exit
            return
        currentScore = self.hand.score
        hasManualScore = self.hasManualScore()
        for box in self.manualRuleBoxes:
            applicable = bool(self.hand.manualRuleMayApply(box.rule))
            if hasManualScore:
                # only those rules which do not affect the score can be applied
                applicable = applicable and box.rule.hasNonValueAction()
            elif box != sender:
                applicable = applicable and self.__ruleChangesScore(
                    box, currentScore)
            box.setApplicable(applicable)

    def __ruleChangesScore(self, box:'RuleBox', currentScore:'Score') ->bool:
        """does the rule actually influence the result?
        if the action would only influence the score and the rule does not change the score,
        ignore the rule. If however the action does other things like penalties leave it applicable"""
        if box.rule.hasNonValueAction():
            return True
        with BlockSignals([box]):
            try:
                checked = box.isChecked()
                box.setChecked(not checked)
                newHand = self.computeHand()
            finally:
                box.setChecked(checked)
        return newHand.score > currentScore

    def __mjstring(self) ->str:
        """compile hand info into a string as needed by the scoring engine"""
        if self.lastTile and self.lastTile.isConcealed:
            lastSource = TileSource.LivingWall.char
        else:
            lastSource = TileSource.LivingWallDiscard.char
        announcements = set()
        rules = [x.rule for x in self.manualRuleBoxes if x.isChecked()]
        for rule in rules:
            options = rule.options
            if 'lastsource' in options:
                if lastSource != TileSource.East14th.char:
                    # this defines precedences for source of last tile
                    lastSource = options['lastsource']
            if 'announcements' in options:
                announcements |= set(options['announcements'])
        return ''.join(['m', lastSource, ''.join(sorted(announcements))])

    def __lastString(self) ->str:
        """compile hand info into a string as needed by the scoring engine"""
        if not self.lastTile:
            return ''
        assert self.handBoard
        if not self.handBoard.tilesByElement(self.lastTile):
            # this happens if we remove the meld with lastTile from the hand
            # again
            return ''
        return f'L{self.lastTile}{self.lastMeld}'

    def computeHand(self) ->Hand:
        """return a Hand object, using a cache"""
        self.lastTile = scoringScene().computeLastTile()
        self.lastMeld = scoringScene().computeLastMeld()
        string = ' '.join(
            [self.scoringString(),
             self.__mjstring(),
             self.__lastString()])
        return Hand(self, string)

    def sortRulesByX(self, rules:List['UsedRule']) ->List['UsedRule']:
        """if this game has a GUI, sort rules by GUI order of the melds they are applied to"""
        rulesWithMelds = [x for x in rules if x.meld]
        rulesWithoutMelds = [x for x in rules if x not in rulesWithMelds]
        assert (handBoard := cast('ScoringHandBoard', self.handBoard))
        tuples:Any = [tuple([x, handBoard.findUIMeld(x.meld)]) for x in rulesWithMelds]
        sorted_tuples = sorted(tuples, key=lambda x: x[1][0].sortKey())
        return [x[0] for x in sorted_tuples] + rulesWithoutMelds

    def addMeld(self, meld:'Meld') ->None:
        """add meld to this hand in a scoring game"""
        if meld.isBonus:
            self._bonusTiles.append(meld[0])
            if Debug.scoring:
                logDebug(f'{self} gets bonus tile {meld[0]}')
        elif meld.isConcealed and not meld.isKong:
            self._concealedMelds.append(meld)
            if Debug.scoring:
                logDebug(f'{self} gets concealed meld {meld}')
        else:
            self._exposedMelds.append(meld)
            if Debug.scoring:
                logDebug(f'{self} gets exposed meld {meld}')
        self._hand = None

    def removeMeld(self, meld:'Meld') ->None:
        """remove a meld from this hand in a scoring game"""
        if meld.isBonus:
            self._bonusTiles.remove(meld[0])
            if Debug.scoring:
                logDebug(f'{self} loses bonus tile {meld[0]}')
        else:
            popped = None
            for melds in [self._concealedMelds, self._exposedMelds]:
                for idx, myMeld in enumerate(melds):
                    if myMeld == meld:
                        popped = melds.pop(idx)
                        break
            if not popped:
                logDebug(
                    f'{self.name}: {self.__class__.__name__}.removeMeld did not find {meld}', showStack=True)
                logDebug(f'    concealed: {self._concealedMelds}')
                logDebug(f'      exposed: {self._exposedMelds}')
            else:
                if Debug.scoring:
                    logDebug(f'{self} lost meld {popped}')
        self._hand = None


class ScoringGame(Game):

    """we play manually on a real table with real tiles and use
    Kajongg only for scoring"""
    playerClass = ScoringPlayer
    wallClass = UIWall

    def __init__(self, names:List[Tuple[Wind, str]], ruleset:'Ruleset', gameid:Optional[int]=None,
        wantedGame:Optional[str]=None, client:Optional['Client']=None) ->None:
        super().__init__(
            names,
            ruleset,
            gameid=gameid,
            client=client,
            wantedGame=wantedGame)
        self.wall:'UIWall'
        self.shouldSave = True
        scoringScene().selectorBoard.load(self)
        self.prepareHand()
        self.initHand()
        scoringScene().mainWindow.adjustMainView()
        scoringScene().mainWindow.updateGUI()
        self.wall.decorate4()
        self.throwDices()

    @Game.seed.getter  # type:ignore[attr-defined]
    def seed(self) ->int: # looks like a pylint bug pylint: disable=invalid-overridden-method
        """a scoring game never has a seed"""
        return 0

    def _setHandSeed(self) ->None:
        """a scoring game does not need this"""
        return None

    def prepareHand(self) ->None:
        """prepare a scoring game hand"""
        Game.prepareHand(self)
        if not self.finished():
            selector = scoringScene().selectorBoard
            selector.refill()
            selector.hasLogicalFocus = True
            self.wall.build(shuffleFirst=False)

    def nextScoringHand(self) ->None:
        """save hand to database, update score table and balance in status line, prepare next hand"""
        if self.winner:
            for player in self.players:
                player.usedDangerousFrom = None
                for ruleBox in player.manualRuleBoxes:
                    rule = ruleBox.rule
                    if rule.name == 'Dangerous Game' and ruleBox.isChecked():
                        self.winner.usedDangerousFrom = player
        self.saveHand()
        self.maybeRotateWinds()
        self.prepareHand()
        self.initHand()
        if _ := scoringScene().scoringDialog:
            _.clear()

    def close(self) ->None:
        """log off from the server and return a Deferred"""
        scoringScene().selectorBoard.uiTiles = []
        scoringScene().selectorBoard.allSelectorTiles = []
        if isAlive(scoringScene()):
            scoringScene().removeTiles()
        for player in self.players:
            player.hide()
        if self.wall:
            self.wall.hide()
        SideText.removeAll()
        return super().close()

    @staticmethod
    def isScoringGame() ->bool:
        """are we scoring a manual game?"""
        return True

    def saveStartTime(self) ->None:
        """write a new entry in the game table with the selected players"""
        Game.saveStartTime(self)
        # for PlayingGame, this one is already done in
        # Connection.__updateServerInfoInDatabase
        known = Query('update server set lastruleset=? where url=?',
                      (self.ruleset.rulesetId, Query.localServerName))
        if not known:
            Query('insert into server(url,lastruleset) values(?,?)',
                  (self.ruleset.rulesetId, Query.localServerName))

    def _setGameId(self) ->None:
        """get a new id"""
        if not self.gameid:
            # a loaded game has gameid already set
            self.gameid = Query("insert into game(seed) values(0)").cursor.lastrowid

    def _mustExchangeSeats(self, pairs:List[List[Player]]) ->List[List[Player]]:
        """filter: which player pairs should really swap places?"""
        # I do not understand the logic of the exec return value. The yes button returns 0
        # and the no button returns 1. According to the C++ doc, the return value is an
        # opaque value that should not be used."""
        return [x for x in pairs if SwapDialog(x).exec() == 0]

    def savePenalty(self, player:'ScoringPlayer', offense:'Rule', amount:int) ->None:
        """save computed values to database, update score table and balance in status line"""
        scoretime = datetime.datetime.now().replace(microsecond=0).isoformat()
        assert self.gameid
        Query("INSERT INTO SCORE "
              "(game,penalty,hand,data,manualrules,player,scoretime,"
              "won,prevailing,wind,points,payments, balance,rotated,notrotated) "
              f"VALUES({int(self.gameid)},1,{int(self.point.handCount)},?,?,{player.nameid},'{scoretime}',"
              f"{int(player == self.winner)},'{self.point.prevailing}','{player.wind}',0,{amount},"
              f"{player.balance},{int(self.point.rotated)}, {int(self.point.notRotated)})",
              (player.hand.string, offense.name))
        assert Internal.mainWindow
        Internal.mainWindow.updateGUI()

def scoreGame() ->Optional[ScoringGame]:
    """show all games, select an existing game or create a new game"""
    Players.load()
    if len(Players.humanNames) < 4:
        logWarning(
            i18n('Please define four players in <interface>Settings|Players</interface>'))
        return None
    gameSelector = Games(Internal.mainWindow)
    selected = None
    if not gameSelector.exec():
        return None
    selected = gameSelector.selectedGame
    gameSelector.close()
    if selected:
        return cast(ScoringGame, ScoringGame.loadFromDB(selected))
    selectDialog = SelectPlayers()
    if not selectDialog.exec():
        return None
    return ScoringGame(list(zip(Wind.all4, selectDialog.names)), selectDialog.cbRuleset.current)
